import pandas as pd
from scipy.sparse import csr_matrix
from sklearn.preprocessing import MultiLabelBinarizer

from immuneML.data_model.EncodedData import EncodedData
from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.encodings.DatasetEncoder import DatasetEncoder
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.util.ParameterValidator import ParameterValidator


class MetadataEncoder(DatasetEncoder):
    """
    Encoder that uses metadata fields as features, such as HLA.

    **Dataset type:**
    - RepertoireDatasets
    - SequenceDatasets
    - ReceptorDatasets

    **Specification arguments:**

    - metadata_fields (list): List of metadata fields to use as features.

    **YAML specification:**

    .. code-block:: yaml

        encodings:
            metadata_encoding:
                Metadata:
                    metadata_fields: [HLA, sex]

    """

    def __init__(self, metadata_fields: list, name: str = None):
        super().__init__(name=name)
        self.metadata_fields = metadata_fields
        self.mlbs = None

    @staticmethod
    def build_object(dataset: Dataset, **params):
        assert 'metadata_fields' in params, "Parameter 'metadata_fields' is required for MetadataEncoder."
        ParameterValidator.assert_type_and_value(params['metadata_fields'], list, 'MetadataEncoder', 'metadata_fields')
        ParameterValidator.assert_all_type_and_value(params['metadata_fields'], str, 'MetadataEncoder', 'metadata_fields')

        ParameterValidator.assert_all_in_valid_list(params['metadata_fields'], dataset.get_label_names(),
                                                   'MetadataEncoder', 'metadata_fields')

        name = params.get('name', 'metadata_encoding')
        return MetadataEncoder(metadata_fields=params['metadata_fields'], name=name)

    def encode(self, dataset, params: EncoderParams) -> Dataset:

        metadata = dataset.get_metadata(self.metadata_fields, return_df=True)

        features = None
        classes = []

        if params.learn_model:
            self.mlbs = {feature: None for feature in self.metadata_fields}

        for feature in self.metadata_fields:
            flattened, mlb = flatten_comma_separated_mlb(metadata, feature, self.mlbs[feature])
            self.mlbs[feature] = mlb
            classes += [f"{feature}_{c}" for c in mlb.classes_.tolist()]
            if features is None:
                features = flattened
            else:
                features = pd.concat([features, flattened], axis=1)

        # Convert to sparse matrix
        onehot_sparse = csr_matrix(features)

        labels = {label: dataset.get_metadata([label])[label] for label in params.label_config.get_labels_by_name()} \
            if params.encode_labels else None

        encoded_dataset = dataset.clone()
        encoded_dataset.encoded_data = EncodedData(
            examples=onehot_sparse,
            feature_names=classes,
            feature_annotations=pd.DataFrame({"feature": classes}),
            labels=labels,
            info={'metadata_fields': self.metadata_fields}
        )

        return encoded_dataset


def flatten_comma_separated_mlb(df, column_name, mlb=None):
    """
    Flatten comma-separated values using MultiLabelBinarizer.

    Parameters:
    -----------
    df : pandas DataFrame
        Input dataframe
    column_name : str
        Name of column containing comma-separated values

    Returns:
    --------
    pandas DataFrame with original columns + one-hot encoded columns
    """

    # Split comma-separated values into lists
    split_values = df[column_name].str.split(',').tolist()

    # Remove whitespace from each element
    split_values = [[item.strip() for item in row if item.strip()] for row in split_values]

    # Create binary matrix
    if mlb is None:
        mlb = MultiLabelBinarizer()
        binary_matrix = mlb.fit_transform(split_values)
    else:
        binary_matrix = mlb.transform(split_values)

    # Create new column names
    new_columns = [f"{column_name}_{label}" for label in mlb.classes_]

    # Create DataFrame with binary columns
    binary_df = pd.DataFrame(binary_matrix, columns=new_columns, index=df.index)

    return binary_df, mlb



from typing import List

import numpy as np
import pandas as pd

from immuneML.data_model.EncodedData import EncodedData
from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.dsl.DefaultParamsLoader import DefaultParamsLoader
from immuneML.encodings.DatasetEncoder import DatasetEncoder
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.util.NumpyHelper import NumpyHelper
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.ReflectionHandler import ReflectionHandler


class CompositeEncoder(DatasetEncoder):
    """
    This encoder allows to combine multiple different encodings together, for example, KmerFrequency encoder
    with VGeneEncoder. The parameters for the different encoders are passed as a list of dictionaries, where each
    dictionary contains the parameters for one encoder. The different encoders are applied sequentially and their
    results concatenated together.

    **Dataset type:**
    - SequenceDatasets
    - ReceptorDatasets
    - RepertoireDatasets

    .. note::

        To combine multiple encodings (e.g., GeneFrequency and KmerFrequency), keep in mind how the ML method will
        use the encoded data downstream. Currently, the recommended way to use CompositeEncoder is with
        :ref:`LogRegressionCustomPenalty`, where you can specify which features should not be penalized.

    **Specification arguments:**

    - encoders (list): A list of dictionaries, where each dictionary contains the parameters for one encoder.

    **YAML specification:**

    .. code-block:: yaml

        encodings:
            my_composite_encoding:
                Composite:
                    encoders:
                        - KmerFrequency:
                            k: 3
                        - GeneFrequency:
                            genes: [V]
                            normalization_type: relative_frequency
                            scale_to_unit_variance: true
                            scale_to_zero_mean: true

    """

    def __init__(self, encoders: List[DatasetEncoder], name: str = None):
        super().__init__(name=name)
        self.encoders = encoders

    @staticmethod
    def build_object(dataset: Dataset, **params):
        assert 'encoders' in params, "Parameter 'encoders' must be provided for CompositeEncoder."
        ParameterValidator.assert_all_type_and_value(params['encoders'], dict, "CompositeEncoder", 'encoders')
        name = params.get('name', 'composite')

        encoders = []
        for step, encoder_specs in enumerate(params['encoders']):
            cls_name = list(encoder_specs.keys())[0] + 'Encoder'
            encoder = ReflectionHandler.get_class_by_name(cls_name, 'encodings')
            default_params = DefaultParamsLoader.load('encodings', cls_name.replace('Encoder', ''))
            encoder_instance = encoder.build_object(dataset, **{**default_params, **encoder_specs[list(encoder_specs.keys())[0]]})
            encoder_instance.name = f"{name}_step_{step+1}_{encoder_instance.name or cls_name}"
            encoders.append(encoder_instance)

        return CompositeEncoder(encoders=encoders, name=name)

    def encode(self, dataset, params: EncoderParams) -> Dataset:
        examples, feature_names, feature_annotations, info = [], [], [], {}

        for encoder in self.encoders:
            encoded_dataset = encoder.encode(dataset, params)
            tmp_feature_annotations = encoded_dataset.encoded_data.feature_annotations
            tmp_feature_annotations['encoder'] = type(encoder).__name__
            tmp_feature_annotations['encoder_name'] = encoder.name
            feature_annotations.append(tmp_feature_annotations)
            examples.append(encoded_dataset.encoded_data.examples)
            feature_names.extend(encoded_dataset.encoded_data.feature_names)
            info[f'encoder_{encoder.name}'] = encoded_dataset.encoded_data.info

        examples = NumpyHelper.concat_arrays_rowwise(examples, use_memmap=True)
        feature_annotations = pd.concat(feature_annotations, ignore_index=True)

        encoded_dataset = dataset.clone()
        encoded_dataset.encoded_data = EncodedData(examples=np.array(examples), feature_names=feature_names,
                                                   feature_annotations=feature_annotations, info=info,
                                                   labels=dataset.get_metadata(params.label_config.get_labels_by_name()),
                                                   encoding=type(self).__name__)

        return encoded_dataset



import json
import shutil
from pathlib import Path

import numpy as np
import pandas as pd
from olga import load_model

from immuneML.data_model.SequenceParams import RegionType, Chain
from immuneML.data_model.bnp_util import write_yaml, read_yaml, get_sequence_field_name
from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.data_model.datasets.ElementDataset import SequenceDataset
from immuneML.environment.SequenceType import SequenceType
from immuneML.ml_methods.generative_models.GenerativeModel import GenerativeModel
from immuneML.ml_methods.generative_models.OLGA import OLGA
from immuneML.util.Logger import print_log
from immuneML.util.PathBuilder import PathBuilder


class SoNNia(GenerativeModel):
    """
    SoNNia models the selection process of T and B cell receptor repertoires. It is based on the SoNNia Python package.
    It supports SequenceDataset as input, but not RepertoireDataset.

    Original publication:
    Isacchini, G., Walczak, A. M., Mora, T., & Nourmohammad, A. (2021). Deep generative selection models of T and B
    cell receptor repertoires with soNNia. Proceedings of the National Academy of Sciences, 118(14), e2023141118.
    https://doi.org/10.1073/pnas.2023141118

    **Specification arguments:**

    - locus (str): The locus of the receptor chain.

    - batch_size (int): number of sequences to use in each batch

    - epochs (int): number of epochs to train the model

    - deep (bool): whether to use a deep model

    - include_joint_genes (bool)

    - n_gen_seqs (int)

    - custom_model_path (str): path for the custom OLGA model if used

    - default_model_name (str): name of the default OLGA model if used

    - seed (int): random seed for the model or None

    - num_processes (int): number of processes to use for sequence generation (default: 4)


     **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            ml_methods:
                my_sonnia_model:
                    SoNNia:
                        batch_size: 1e4
                        epochs: 5
                        default_model_name: humanTRB
                        deep: False
                        include_joint_genes: True
                        n_gen_seqs: 100

    """

    @classmethod
    def load_model(cls, path: Path):
        from sonnia.sonnia import SoNNia as InternalSoNNia

        assert path.exists(), f"{cls.__name__}: {path} does not exist."

        model_overview_file = path / 'model_overview.yaml'

        for file in [model_overview_file]:
            assert file.exists(), f"{cls.__name__}: {file} is not a file."

        model_overview = read_yaml(model_overview_file)
        sonnia = SoNNia(**{k: v for k, v in model_overview.items() if k != 'type'})
        with open(path / 'model.json', 'r') as json_file:
            model_data = json.load(json_file)

        sonnia._model = InternalSoNNia(custom_pgen_model=sonnia._model_path, seed=sonnia.seed,
                                       vj=sonnia.locus in [Chain.ALPHA, Chain.KAPPA, Chain.LIGHT],
                                       include_joint_genes=sonnia.include_joint_genes,
                                       include_indep_genes=not sonnia.include_joint_genes)

        sonnia._model.model.set_weights([np.array(w) for w in model_data['model_weights']])

        return sonnia

    def __init__(self, locus=None, batch_size: int = None, epochs: int = None, deep: bool = False, name: str = None,
                 default_model_name: str = None, n_gen_seqs: int = None, include_joint_genes: bool = True,
                 custom_model_path: str = None, seed: int = None, num_processes: int = 1):

        if locus is not None:
            super().__init__(Chain.get_chain(str(locus)), region_type=RegionType.IMGT_JUNCTION, seed=seed)
        elif default_model_name is not None:
            super().__init__(locus=Chain.get_chain(default_model_name[-3:]), region_type=RegionType.IMGT_JUNCTION,
                             seed=seed)
        else:
            super().__init__(locus=None, region_type=RegionType.IMGT_JUNCTION, seed=seed)
        self.epochs = epochs
        self.batch_size = int(batch_size)
        self.deep = deep
        self.include_joint_genes = include_joint_genes
        self.n_gen_seqs = n_gen_seqs
        self._model = None
        self.name = name
        self.num_processes = max(num_processes, 1)
        self.default_model_name = default_model_name
        self._model_path = custom_model_path

        self.set_model_path(self.default_model_name)

    def set_model_path(self, model_name: str):
        if self._model_path is None or self._model_path == '':
            if model_name is not None:
                self.default_model_name = model_name
                self._model_path = Path(
                    load_model.__file__).parent / f"default_models/{OLGA.DEFAULT_MODEL_FOLDER_MAP[self.default_model_name]}"

    def fit(self, dataset: Dataset, path: Path = None):
        from sonnia.sonnia import SoNNia as InternalSoNNia

        if self._model_path is None:
            self.set_locus(dataset)
            self.set_model_path(f"human{self.locus.value}")

        print_log(f"{SoNNia.__name__}: fitting a selection model for {self.default_model_name}...", True)

        data = dataset.data.topandas()[['junction_aa', 'v_call', 'j_call']]
        data_seqs = data.to_records(index=False).tolist()

        self._model = InternalSoNNia(data_seqs=data_seqs,
                                     gen_seqs=[],
                                     custom_pgen_model=self._model_path,
                                     vj=self.locus in [Chain.ALPHA, Chain.KAPPA, Chain.LIGHT],
                                     include_joint_genes=self.include_joint_genes,
                                     include_indep_genes=not self.include_joint_genes)

        self._model.add_generated_seqs(num_gen_seqs=self.n_gen_seqs, custom_model_folder=self._model_path,
                                       processes=self.num_processes)

        self._model.infer_selection(epochs=self.epochs, batch_size=self.batch_size, verbose=1)

        print_log(f"{SoNNia.__name__}: selection model fitted.", True)

    def is_same(self, model) -> bool:
        raise NotImplementedError

    def generate_sequences(self, count: int, seed: int, path: Path, sequence_type: SequenceType, compute_p_gen: bool):
        from sonia.sequence_generation import SequenceGeneration

        gen_model = SequenceGeneration(self._model, processes=self.num_processes)
        sequences = gen_model.generate_sequences_post(count)

        df = pd.DataFrame({
            get_sequence_field_name(self.region_type, SequenceType.AMINO_ACID): [seq[0] for seq in sequences],
            get_sequence_field_name(self.region_type, SequenceType.NUCLEOTIDE): [seq[3] for seq in sequences],
            'v_call': [seq[1] for seq in sequences],
            'j_call': [seq[2] for seq in sequences],
            'gen_model_name': [self.name] * count,
            'locus': [self.locus.to_string()] * count
        })

        return SequenceDataset.build_from_partial_df(df, path=PathBuilder.build(path), name='SoNNiaDataset',
                                                     labels={'gen_model_name': [self.name]},
                                                     type_dict={'gen_model_name': str})

    def compute_p_gens(self, sequences, sequence_type: SequenceType) -> np.ndarray:
        raise NotImplementedError

    def compute_p_gen(self, sequence: dict, sequence_type: SequenceType) -> float:
        raise NotImplementedError

    def can_compute_p_gens(self) -> bool:
        return False

    def can_generate_from_skewed_gene_models(self) -> bool:
        return False

    def generate_from_skewed_gene_models(self, v_genes: list, j_genes: list, seed: int, path: Path,
                                         sequence_type: SequenceType, batch_size: int, compute_p_gen: bool):
        raise NotImplementedError

    def save_model(self, path: Path) -> Path:
        PathBuilder.build(path / 'model')

        write_yaml(path / 'model/model_overview.yaml', {'type': 'SoNNia', 'locus': self.locus.name,
                                                        **{k: v for k, v in vars(self).items()
                                                           if
                                                           k not in ['_model', 'locus', '_model_path', 'region_type']}})
        attributes_to_save = ['data_seqs', 'gen_seqs', 'log']
        self._model.save_model(path / 'model', attributes_to_save)

        model_json = self._model.model.to_json()
        model_weights = [w.tolist() for w in self._model.model.get_weights()]
        model_data = {'model_config': model_json, 'model_weights': model_weights}
        with open(path / 'model' / 'model.json', 'w') as json_file:
            json.dump(model_data, json_file)

        return Path(shutil.make_archive(str(path / f'trained_model_{self.name}'), "zip", str(path / 'model'))).absolute()



import logging
from pathlib import Path

import pandas as pd
import plotly.express as px

from immuneML.environment.Label import Label
from immuneML.hyperparameter_optimization.states.TrainMLModelState import TrainMLModelState
from immuneML.reports.PlotlyUtil import PlotlyUtil
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.train_ml_model_reports.TrainMLModelReport import TrainMLModelReport
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder


class MLSettingsPerformance(TrainMLModelReport):
    """
    Report for TrainMLModel instruction: plots the performance for each of the setting combinations as defined under 'settings' in the
    assessment (outer validation) loop.

    The performances are grouped by label (horizontal panels) encoding (vertical panels) and ML method (bar color).
    When multiple data splits are used, the average performance over the data splits is shown with an error bar
    representing the standard deviation.

    This report can be used only with TrainMLModel instruction under 'reports'.


    **Specification arguments:**

    - single_axis_labels (bool): whether to use single axis labels. Note that using single axis labels makes the figure unsuited for rescaling, as the label position is given in a fixed distance from the axis. By default, single_axis_labels is False, resulting in standard plotly axis labels.

    - x_label_position (float): if single_axis_labels is True, this should be an integer specifying the x axis label position relative to the x axis. The default value for label_position is -0.1.

    - y_label_position (float): same as x_label_position, but for the y-axis.


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            reports:
                my_hp_report: MLSettingsPerformance

    """

    @classmethod
    def build_object(cls, **kwargs):
        location = "MLSettingsPerformance"

        single_axis_labels = kwargs["single_axis_labels"]
        ParameterValidator.assert_type_and_value(single_axis_labels, bool, location, "single_axis_labels")

        if single_axis_labels:
            x_label_position = kwargs["x_label_position"]
            ParameterValidator.assert_type_and_value(x_label_position, float, location, "x_label_position")
            y_label_position = kwargs["y_label_position"]
            ParameterValidator.assert_type_and_value(y_label_position, float, location, "y_label_position")
        else:
            x_label_position = None
            y_label_position = None

        name = kwargs["name"] if "name" in kwargs else None
        return MLSettingsPerformance(single_axis_labels, x_label_position, y_label_position, name)

    def __init__(self, single_axis_labels, x_label_position, y_label_position, name: str = None, state: TrainMLModelState = None,
                 label: Label = None, result_path: Path = None, number_of_processes: int = 1):
        super().__init__(name=name, state=state, label=label, result_path=result_path, number_of_processes=number_of_processes)
        self.single_axis_labels = single_axis_labels
        self.x_label_position = x_label_position
        self.y_label_position = y_label_position
        self.result_name = "performance"
        self.vertical_grouping = "encoding"

    def _generate(self) -> ReportResult:
        PathBuilder.build(self.result_path)

        plotting_data = self._retrieve_plotting_data()
        result_table = self._write_results_table(plotting_data)
        report_output_fig = self._safe_plot(plotting_data=plotting_data)
        output_figures = [report_output_fig] if report_output_fig is not None else []

        return ReportResult(self.name,
                            info="The performance on the test set in the assessment (outer cross-validation) loop for each of the setting combinations as defined under 'settings'.",
                            output_tables=[result_table], output_figures=output_figures)

    def _get_vertical_grouping(self, assessment_item):
        return assessment_item.hp_setting.encoder_name

    def _get_color_grouping(self, assessment_item):
        return assessment_item.hp_setting.ml_method_name

    def _retrieve_plotting_data(self):
        plotting_data = []

        for assessment_state in self.state.assessment_states:
            for label_key, label_state in assessment_state.label_states.items():
                for assessment_key, assessment_item in label_state.assessment_items.items():
                    plotting_data.append([assessment_state.split_index + 1,
                                          label_key,
                                          self._get_vertical_grouping(assessment_item),
                                          self._get_color_grouping(assessment_item),
                                          assessment_item.performance[self.state.optimization_metric.name.lower()]])
                    # optional: include assessment_item.hp_setting.preproc_sequence_name. for now ignored.

        plotting_data = pd.DataFrame(plotting_data, columns=["split", "label", self.vertical_grouping, "ml_method", "performance"])
        plotting_data.replace(to_replace=[None], value="NA", inplace=True)

        return plotting_data

    def _write_results_table(self, plotting_data):
        filepath = self.result_path / f"{self.result_name}.csv"
        plotting_data.to_csv(filepath, index=False)
        return ReportOutput(filepath, name=f"Performance table per encoding and ML method across CV splits")

    def std(self, x):
        return x.std(ddof=0)

    def _plot(self, plotting_data):
        plotting_data = self._preprocess_plotting_data(plotting_data)

        metric_name = self.state.optimization_metric.name.replace("_", " ").title()

        if self.single_axis_labels:
            figure = self._plot_single_axis_labels(plotting_data, "ML method", f"Performance ({metric_name})")
        else:
            figure = self._plot_rescalable(plotting_data, "ML method", f"Performance<br>({metric_name})")

        file_path = self.result_path / f"{self.result_name}.html"
        file_path = PlotlyUtil.write_image_to_file(figure, file_path)

        return ReportOutput(path=file_path)

    def _preprocess_plotting_data(self, plotting_data):
        plotting_data = plotting_data.groupby(["label", self.vertical_grouping, "ml_method"], as_index=False).agg(
            {"split": "first", "performance": ['mean', self.std]})

        plotting_data.columns = plotting_data.columns.map(''.join)

        return plotting_data

    def _plot_rescalable(self, plotting_data, x_label, y_label):
        figure = px.bar(plotting_data, x="ml_method", y="performancemean", color="ml_method", barmode="relative",
                        facet_row=self.vertical_grouping, facet_col="label", error_y="performancestd",
                        labels={
                            "performancemean": y_label,
                            "ml_method": x_label,
                        }, template='plotly_white',
                        color_discrete_sequence=px.colors.diverging.Tealrose)
        figure.update_layout(showlegend=False)
        return figure

    def _plot_single_axis_labels(self, plotting_data, x_label, y_label):
        figure = self._plot_rescalable(plotting_data, x_label, y_label)
        return PlotlyUtil.add_single_axis_labels(figure, x_label, y_label, self.x_label_position, self.y_label_position)

    def check_prerequisites(self):
        run_report = True

        if self.state is None:
            logging.warning(f"{self.__class__.__name__} can only be executed as a hyperparameter report. MLSettingsPerformance report will not be created.")
            run_report = False

        if self.result_path is None:
            logging.warning(f"{self.__class__.__name__} requires an output 'path' to be set. MLSettingsPerformance report will not be created.")
            run_report = False

        return run_report



from immuneML.encodings.DatasetEncoder import DatasetEncoder
from immuneML.ml_methods.classifiers.MLMethod import MLMethod


class HPSetting:

    def __init__(self, encoder: DatasetEncoder, encoder_params: dict,
                 ml_method: MLMethod, ml_params: dict,
                 preproc_sequence: list, encoder_name: str = None,
                 ml_method_name: str = None, preproc_sequence_name: str = None):

        self.encoder = encoder
        self.encoder_params = encoder_params
        self.ml_method = ml_method
        self.ml_params = ml_params
        self.preproc_sequence = preproc_sequence
        self.encoder_name = encoder_name
        self.ml_method_name = ml_method_name
        self.preproc_sequence_name = preproc_sequence_name

    def get_key(self):
        key = f"{self.encoder_name}_{self.ml_method_name}"
        if self.preproc_sequence is not None and self.preproc_sequence_name is not None:
            key += f"_{self.preproc_sequence_name}"
        return key

    def __str__(self):
        return self.get_key()



from pathlib import Path

import pystache


class TemplateParser:

    @staticmethod
    def parse(template_path: Path, template_map: dict, result_path: Path) -> Path:
        with template_path.open("r") as template_file:
            template = template_file.read()

        rendered_template = pystache.render(template, template_map)

        with result_path.open("w") as file:
            file.writelines(rendered_template)

        return result_path



import argparse
import logging
import os
import warnings
from pathlib import Path

from immuneML.dsl.ImmuneMLParser import ImmuneMLParser
from immuneML.dsl.semantic_model.SemanticModel import SemanticModel
from immuneML.dsl.symbol_table.SymbolType import SymbolType
from immuneML.util.Logger import print_log
from immuneML.util.PathBuilder import PathBuilder


class SimError(Exception):
    pass


class LigoApp:

    def __init__(self, specification_path: Path, result_path: Path):
        self._specification_path = Path(specification_path)
        self._result_path = Path(os.path.relpath(result_path))

        PathBuilder.build(self._result_path)

    def run(self):
        try:
            print_log(f"LIgO: parsing the specification...\n", include_datetime=True)

            symbol_table, self._specification_path = ImmuneMLParser.parse_yaml_file(self._specification_path,
                                                                                    self._result_path)

            print_log(f"LIgO: starting the simulation...\n", include_datetime=True)

            instructions = symbol_table.get_by_type(SymbolType.INSTRUCTION)
            output = symbol_table.get("output")
            model = SemanticModel([instruction.item for instruction in instructions], self._result_path, output)
            result = model.run()

            print_log(f"LIgO: finished simulation.\n", include_datetime=True)

            return result
        except SimError as e:
            print(e)


def run_ligo(namespace: argparse.Namespace):
    if os.path.isdir(namespace.result_path) and len(os.listdir(namespace.result_path)) != 0:
        raise ValueError(
            f"Directory {namespace.result_path} already exists. Please specify a new output directory for the analysis.")
    PathBuilder.build(namespace.result_path)

    logging.basicConfig(filename=Path(namespace.result_path) / "log.txt", level=logging.INFO,
                        format='%(asctime)s %(levelname)s: %(message)s')
    warnings.showwarning = lambda message, category, filename, lineno, file=None, line=None: logging.warning(message)

    app = LigoApp(namespace.specification_path, namespace.result_path)
    app.run()


def main():
    parser = argparse.ArgumentParser(description="LIgO command line tool")
    parser.add_argument("specification_path",
                        help="Path to specification YAML file. Always used to define the simulation.")
    parser.add_argument("result_path", help="Output directory path.")

    namespace = parser.parse_args()
    namespace.specification_path = Path(namespace.specification_path)
    namespace.result_path = Path(namespace.result_path)

    run_ligo(namespace)


if __name__ == "__main__":
    main()



import argparse
import glob
from pathlib import Path

from immuneML.app.ImmuneMLApp import ImmuneMLApp


def run_all_yamls(directory_path, result_path):

    yaml_files = list(glob.glob(f"{Path(directory_path)}/*.yaml"))

    if not yaml_files:
        print("No YAML files found in the specified directory.")
        return

    for yaml_file in yaml_files:
        yaml_path = Path(yaml_file)
        run_result_path = Path(result_path) / f"{yaml_path.stem}_result"
        print(f"\nProcessing {yaml_file}...")

        try:
            ImmuneMLApp(specification_path=yaml_path, result_path=run_result_path).run()
            print(f"Successfully completed {yaml_file}")
            print(f"Results saved to: {run_result_path}")
        except Exception as e:
            print(f"Error processing {yaml_file}: {str(e)}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Run all YAML files in a directory using immuneML')
    parser.add_argument('--directory', '-d', default=".",
                        help='Directory containing YAML files (default: current directory)')
    parser.add_argument('--result-path', '-r', default="./results",
                        help='Directory to store results (default: ./results)')

    args = parser.parse_args()
    run_all_yamls(args.directory, args.result_path)



import dataclasses
import logging
from dataclasses import fields as get_fields
from enum import Enum
from itertools import chain
from pathlib import Path
from typing import Dict, Any, Tuple, List

import bionumpy as bnp
import numpy as np
import pandas as pd
import yaml
from bionumpy.bnpdataclass import bnpdataclass
from bionumpy.encoded_array import EncodedArray

from immuneML.data_model.AIRRSequenceSet import AIRRSequenceSet
from immuneML.data_model.SequenceParams import Chain, RegionType
from immuneML.environment.SequenceType import SequenceType
from immuneML.util.ReflectionHandler import ReflectionHandler


def bnp_write_to_file(filename: Path, bnp_object):
    buffer_type = bnp.io.delimited_buffers.get_bufferclass_for_datatype(type(bnp_object), delimiter='\t',
                                                                        has_header=True)
    with bnp.open(str(filename), 'w', buffer_type=buffer_type) as file:
        file.write(bnp_object)


def bnp_read_from_file(filename: Path, buffer_type: bnp.io.delimited_buffers.DelimitedBuffer = None, dataclass=None):
    if buffer_type is None:
        buffer_type = bnp.io.delimited_buffers.get_bufferclass_for_datatype(dataclass, delimiter='\t', has_header=True)
    with bnp.open(str(filename), buffer_type=buffer_type) as file:
        return file.read()  # TODO: fix - throws error when empty file (no lines after header)


def write_dataset_yaml(filename: Path, yaml_dict):
    for mandatory_field in ["identifier", "dataset_type", "name", "labels"]:
        assert mandatory_field in yaml_dict.keys(), f"Error exporting {filename.stem}: missing field {mandatory_field}"

    if yaml_dict["dataset_type"] == "RepertoireDataset":
        assert "metadata_file" in yaml_dict.keys(), f"Error exporting {filename.stem}: missing field metadata_file"

    if yaml_dict["dataset_type"] in ("SequenceDataset", "ReceptorDataset"):
        assert "filename" in yaml_dict.keys(), f"Error exporting {filename.stem}: missing field filename"
        assert "type_dict_dynamic_fields" in yaml_dict.keys(), f"Error exporting {filename.stem}: missing field type_dict_dynamic_fields"

    assert type(yaml_dict["labels"]) == dict or type(yaml_dict["labels"]) == None, "labels format must be dict or None"

    write_yaml(filename, yaml_dict)


def write_yaml(filename: Path, yaml_dict):
    with filename.open('w') as file:
        yaml.dump({k: str(v) if isinstance(v, Path) else v for k, v in yaml_dict.items()}, file)
    return filename


def read_yaml(filename: Path) -> dict:
    with filename.open("r") as file:
        content = yaml.safe_load(file)
    return load_type_dict(content)


def get_sequence_field_name(region_type: RegionType = RegionType.IMGT_CDR3,
                            sequence_type: SequenceType = SequenceType.AMINO_ACID):
    suffix = "_aa" if sequence_type == SequenceType.AMINO_ACID else ""
    return region_type.value + suffix


def load_type_dict(full_dict: dict) -> dict:
    if 'type_dict' in full_dict:
        full_dict['type_dict'] = {key: AIRRSequenceSet.STR_TO_TYPE[val] for key, val in full_dict['type_dict'].items()}
    return full_dict


def build_dynamic_bnp_dataclass(all_fields_dict: Dict[str, Any]):
    sequence_field_names = {field.name: field.type for field in dataclasses.fields(AIRRSequenceSet)}
    types = {}

    for key, value in all_fields_dict.items():
        if key in sequence_field_names:
            field_type = sequence_field_names[key]
        else:
            field_type = get_field_type_from_values(value)
        types[key] = field_type

    dc = AIRRSequenceSet.extend(tuple((name, t, AIRRSequenceSet.get_neutral_value(t)) for name, t in types.items()
                                      if name not in list(AIRRSequenceSet.get_field_type_dict().keys())))
    return dc, types


def build_dynamic_bnp_dataclass_obj(all_fields_dict: Dict[str, Any]):
    dc, types = build_dynamic_bnp_dataclass(all_fields_dict)
    all_fields_dict = add_neutral_values(all_fields_dict, types)
    all_fields_dict = convert_to_expected_types(all_fields_dict, types)
    all_fields_dict = make_full_airr_seq_set_df(pd.DataFrame(all_fields_dict)).to_dict(orient='list')
    return dc(**all_fields_dict), types


def convert_to_expected_types(all_fields_dict, types) -> dict:
    for field_name, field_type in types.items():
        all_fields_dict[field_name] = [field_type(val) for val in all_fields_dict[field_name]]

    return all_fields_dict


def prepare_values_for_bnp(field_values: dict, types: dict) -> dict:
    values = add_neutral_values(field_values, types)
    values = convert_enums_to_str(values)
    return values


def add_neutral_values(field_values: dict, types: dict) -> dict:
    return {
        field: [val if val is not None else AIRRSequenceSet.get_neutral_value(types[field]) for val in values]
        for field, values in field_values.items()
    }


def convert_enums_to_str(field_values: dict) -> dict:
    return {
        field: [val.to_string() if isinstance(val, Enum) and hasattr(val, 'to_string') else val for val in values]
        for field, values in field_values.items()
    }


def extend_dataclass_with_dynamic_fields(cls, fields: List[Tuple[str, type]], cls_name: str = None):
    cls_name = cls_name if cls_name is not None else "Dynamic" + cls.__name__
    new_cls = bnpdataclass(dataclasses.make_dataclass(cls_name,
                                                      fields=[(name, field_type, dataclasses.field(default=None)) for
                                                              name, field_type in fields], bases=(cls,)))

    def dynamic_fields(cls):
        return [el[0] for el in fields]

    new_cls.dynamic_fields = classmethod(dynamic_fields)

    return new_cls


def get_field_type_from_values(values):
    if isinstance(values, np.ndarray):
        t = type(values[0].item())
    elif len(values) == 0:
        t = str
    elif values[0] is not None:
        if issubclass(type(values[0]), Enum):
            t = str
        else:
            t = type(values[0])
    else:
        proper_values = [v for v in values if v is not None]
        if len(proper_values) > 0:
            t = type(proper_values[0])
        else:
            t = str

    return t


def get_row_by_index(self, index) -> dict:
    field_names = [f.name for f in dataclasses.fields(self)]
    d = dict()

    for field_name in field_names:
        field_value = getattr(self, field_name)[index]

        if isinstance(field_value, EncodedArray):
            field_value = field_value.to_string()
        else:
            field_value = field_value.item()

        d[field_name] = field_value

    return d


def get_rows_by_indices(self, index1, index2) -> dict:
    row1 = self.get_row_by_index(index1)
    row2 = self.get_row_by_index(index2)

    assert row1['cell_id'] == row2['cell_id'], (row1['cell_id'], row2['cell_id'])

    return {
        **{f'{Chain.get_chain(row1["locus"]).name.lower()}_{key}': val for key, val in row1.items()},
        **{f'{Chain.get_chain(row2["locus"]).name.lower()}_{key}': val for key, val in row2.items()},
    }


def get_single_row_value(self, attr_name: str):
    if hasattr(self, attr_name) and getattr(self, attr_name) is not None:
        val = getattr(self, attr_name)
        if isinstance(val, EncodedArray):
            return val.to_string()
        elif isinstance(val, np.ndarray):
            return val.item()
    else:
        return None


def make_element_dataset_objects(bnp_object, class_name) -> list:
    cls = ReflectionHandler.get_class_by_name(class_name, 'data_model')
    if class_name == 'ReceptorSequence':
        objects = [cls.create_from_record(**bnp_object.get_row_by_index(i)) for i in range(len(bnp_object))]
    else:
        objects = [cls.create_from_record(**bnp_object.get_rows_by_indices(i, i + 1)) for i in
                   range(0, len(bnp_object), 2)]
    return objects


def make_buffer_type_from_dataset_file(dataset_file: Path):
    if dataset_file is not None and dataset_file.exists() and dataset_file.is_file():
        with dataset_file.open('r') as file:
            metadata = yaml.safe_load(file)

        type_dict = {key: AIRRSequenceSet.STR_TO_TYPE[val] for key, val in metadata["type_dict"].items()}
        dataclass = extend_dataclass_with_dynamic_fields(AIRRSequenceSet, list(type_dict.items()))
        return bnp.io.delimited_buffers.get_bufferclass_for_datatype(dataclass, delimiter='\t', has_header=True)
    else:
        raise RuntimeError(f"Dataset file {dataset_file} doesn't exist, cannot load the dataset.")


def merge_dataclass_objects(objects: list, fill_unmatched: bool = False):
    fields = {k: v for d in [{field.name: field.type for field in get_fields(obj)} for obj in objects] for k, v in
              d.items()}
    fields = {k: fields[k] for k in sorted(list(fields.keys()))}

    tmp_objs = []

    for obj in objects:
        missing_fields = [field for field in fields.keys() if not hasattr(obj, field)]

        if not fill_unmatched or len(missing_fields) == 0:
            assert all(hasattr(obj, field) for field in fields.keys()), (obj, list(fields.keys()))
            tmp_objs.append(obj)
        else:
            logging.info(f"Filling in missing fields when merging dataclass objects: {missing_fields}\nObject:\n{obj}")
            tmp_objs.append(
                obj.add_fields({field_name: [AIRRSequenceSet.get_neutral_value(fields[field_name])] * len(obj)
                                for field_name in missing_fields},
                               {field_name: field_type for field_name, field_type in fields.items() if
                                field_name in missing_fields}))

    cls = type(tmp_objs[0])
    return cls(
        **{field_name: list(chain.from_iterable([getattr(obj, field_name) for obj in tmp_objs])) for field_name in
           fields.keys()})


def get_type_dict_from_bnp_object(bnp_object) -> dict:
    return {field.name: field.type for field in get_fields(bnp_object)}


def make_full_airr_seq_set_df(df):
    field_type_dict = AIRRSequenceSet.get_field_type_dict()
    default_fields = pd.DataFrame({
        f_name: [AIRRSequenceSet.get_neutral_value(f_type) for _ in range(df.shape[0])]
        for f_name, f_type in field_type_dict.items() if f_name not in df.columns})
    return pd.concat([df, default_fields], axis=1)



import inspect
from pathlib import Path

from immuneML.dsl.symbol_table.SymbolTable import SymbolTable
from immuneML.dsl.symbol_table.SymbolType import SymbolType
from immuneML.hyperparameter_optimization.config.ManualSplitConfig import ManualSplitConfig
from immuneML.hyperparameter_optimization.config.SplitType import SplitType
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.workflows.instructions.train_gen_model.TrainGenModelInstruction import TrainGenModelInstruction


class TrainGenModelParser:

    def parse(self, key: str, instruction: dict, symbol_table: SymbolTable, path: Path = None) -> TrainGenModelInstruction:
        valid_keys = [k for k in inspect.signature(TrainGenModelInstruction.__init__).parameters.keys()
                      if k not in ['result_path', 'name', 'self']] + ['type', 'method']

        ParameterValidator.assert_all_in_valid_list(list(instruction.keys()), valid_keys, TrainGenModelParser.__name__, key)

        dataset = symbol_table.get(instruction['dataset'])

        assert ("method" in instruction) ^ ("methods" in instruction), \
            f"{TrainGenModelParser.__name__}: either 'method' or 'methods' key must be specified in the instruction."

        if 'method' in instruction:
            methods = [symbol_table.get(instruction['method'])]
        else:
            methods = [symbol_table.get(method_id) for method_id in instruction['methods']]

        assert type(dataset).__name__ in ['SequenceDataset', 'ReceptorDataset'], \
            (f'{TrainGenModelParser.__name__}: TrainGenModel instruction can for now be used only with '
             f'Sequence/Receptor datasets; Repertoire datasets are not supported.')

        for method in methods:
            assert method.__class__.__name__ not in ['ExperimentalImport', 'OLGA'], \
                (f"{TrainGenModelParser.__name__}: ExperimentalImport and OLGA cannot be used with TrainGenModel "
                 f"instruction. Please specify some of the other generative models.")

        ParameterValidator.assert_type_and_value(instruction['gen_examples_count'], int, TrainGenModelParser.__name__,
                                                 'gen_examples_count', 0)
        ParameterValidator.assert_type_and_value(instruction['number_of_processes'], int, TrainGenModelParser.__name__,
                                                 'number_of_processes', 1)
        ParameterValidator.assert_type_and_value(float(instruction['training_percentage']), float,
                                                 TrainGenModelParser.__name__,
                                                 'training_percentage', 0, 1)
        ParameterValidator.assert_type_and_value(instruction['export_generated_dataset'], bool,
                                                 TrainGenModelParser.__name__, 'export_generated_dataset')
        ParameterValidator.assert_type_and_value(instruction['export_combined_dataset'], bool,
                                                 TrainGenModelParser.__name__, 'export_combined_dataset')
        ParameterValidator.assert_type_and_value(instruction['split_strategy'], str, TrainGenModelParser.__name__, 'split_strategy')
        ParameterValidator.assert_type_and_value(instruction['split_config'], dict, TrainGenModelParser.__name__, 'split_config', nullable=True)

        assert instruction['export_generated_dataset'] or instruction['export_combined_dataset'], \
            (f"{TrainGenModelParser.__name__}: 'export_generated_dataset' and 'export_combined_dataset' are both set "
             f"to False. At least one of these must be True. ")

        valid_report_ids = symbol_table.get_keys_by_type(SymbolType.REPORT)
        ParameterValidator.assert_all_in_valid_list(instruction['reports'], valid_report_ids, TrainGenModelParser.__name__, 'reports')

        reports = [symbol_table.get(report_id) for report_id in instruction['reports']]

        if instruction['split_strategy'].upper() == SplitType.RANDOM.name:
            split_strategy = SplitType.RANDOM
            split_config = None
        elif instruction['split_strategy'].upper() == SplitType.MANUAL.name:
            split_strategy = SplitType.MANUAL
            split_config = ManualSplitConfig(**instruction['split_config'])
        else:
            raise ValueError(f"{TrainGenModelParser.__name__}: Unsupported split strategy '{instruction['split_strategy']}'.")

        return TrainGenModelInstruction(**{**{k: v for k, v in instruction.items() if k not in ['type', 'method']},
                                           **{'dataset': dataset, 'methods': methods, 'name': key, 'reports': reports,
                                              'split_strategy': split_strategy, 'split_config': split_config}})



from typing import List

import numpy as np
import pandas as pd

from immuneML.analysis.SequenceMatcher import SequenceMatcher
from immuneML.caching.CacheHandler import CacheHandler
from immuneML.data_model.SequenceParams import Chain
from immuneML.data_model.datasets.RepertoireDataset import RepertoireDataset
from immuneML.data_model.EncodedData import EncodedData
from immuneML.data_model.SequenceSet import Receptor
from immuneML.data_model.SequenceSet import Repertoire
from immuneML.encodings.DatasetEncoder import DatasetEncoder
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.encodings.reference_encoding.MatchedReferenceUtil import MatchedReferenceUtil
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.ReadsType import ReadsType


class MatchedReceptorsEncoder(DatasetEncoder):
    """
    Encodes the dataset based on the matches between a dataset containing unpaired (single chain) data,
    and a paired reference receptor dataset.
    For each paired reference receptor, the frequency of either chain in the dataset is counted.

    This encoding can be used in combination with the :ref:`Matches` report.

    When sum_matches and normalize are set to True, this encoder behaves similarly as described in: Yao, Y. et al. ‘T cell receptor repertoire as a potential diagnostic marker for celiac disease’.
    Clinical Immunology Volume 222 (January 2021): 108621. `doi.org/10.1016/j.clim.2020.108621 <https://doi.org/10.1016/j.clim.2020.108621>`_
    with the only exception being that this encoder uses paired receptors, while the original publication used single sequences (see also: :ref:`MatchedSequences` encoder).


    **Dataset type:**

    - RepertoireDatasets


    **Specification arguments:**

    - reference (dict): A dictionary describing the reference dataset file. Import should be specified the same way as
      regular dataset import. It is only allowed to import a receptor dataset here (i.e., is_repertoire is False and
      paired is True by default, and these are not allowed to be changed).

    - max_edit_distances (dict): A dictionary specifying the maximum edit distance between a target sequence (from the
      repertoire) and the reference sequence. A maximum distance can be specified per chain, for example to allow for
      less strict matching of TCR alpha and BCR light chains. When only an integer is specified, this distance is
      applied to all possible chains.

    - reads (:py:mod:`~immuneML.util.ReadsType`): Reads type signify whether the counts of the sequences in the
      repertoire will be taken into account. If :py:mod:`~immuneML.util.ReadsType.UNIQUE`, only unique sequences
      (clonotypes) are counted, and if :py:mod:`~immuneML.util.ReadsType.ALL`, the sequence 'count' value is summed when
      determining the number of matches. The default value for reads is all.

    - sum_matches (bool): When sum_matches is False, the resulting encoded data matrix contains multiple columns with
      the number of matches per reference receptor chain. When sum_matches is true, the columns representing each of the
      two chains are summed together, meaning that there are only two aggregated sums of matches (one per chain) per
      repertoire in the encoded data. To use this encoder in combination with the :ref:`Matches` report, sum_matches
      must be set to False. When sum_matches is set to True, this encoder behaves similarly to the encoder described by
      Yao, Y. et al. By default, sum_matches is False.

    - normalize (bool): If True, the chain matches are divided by the total number of unique receptors in the repertoire
      (when reads = unique) or the total number of reads in the repertoire (when reads = all).


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            encodings:
                my_mr_encoding:
                    MatchedReceptors:
                        reference:
                            format: VDJdb
                            params:
                                path: path/to/file.txt
                        max_edit_distances:
                            TRA: 1
                            TRB: 0
    """

    dataset_mapping = {
        "RepertoireDataset": "MatchedReceptorsRepertoireEncoder"
    }

    def __init__(self, reference: List[Receptor], max_edit_distances: dict, reads: ReadsType, sum_matches: bool,
                 normalize: bool, name: str = None):
        super().__init__(name=name)
        self.reference_receptors = reference
        self.max_edit_distances = max_edit_distances
        self.reads = reads
        self.sum_matches = sum_matches
        self.normalize = normalize
        self.feature_count = 2 if self.sum_matches else len(self.reference_receptors) * 2

    @staticmethod
    def _prepare_parameters(reference: dict, max_edit_distances: dict, reads: str, sum_matches: bool, normalize: bool,
                            name: str = None):
        location = "MatchedReceptorsEncoder"

        ParameterValidator.assert_type_and_value(sum_matches, bool, location, "sum_matches")
        ParameterValidator.assert_type_and_value(normalize, bool, location, "normalize")
        ParameterValidator.assert_in_valid_list(reads.upper(), [item.name for item in ReadsType], location, "reads")

        legal_chains = [chain.value for chain in Chain]

        if type(max_edit_distances) is int:
            max_edit_distances = {chain: max_edit_distances for chain in legal_chains}
        elif type(max_edit_distances) is dict:
            ParameterValidator.assert_keys(max_edit_distances.keys(), legal_chains, location, "max_edit_distances",
                                           exclusive=False)
        else:
            ParameterValidator.assert_type_and_value(max_edit_distances, dict, location, 'max_edit_distances')

        reference_receptors = MatchedReferenceUtil.prepare_reference(reference, location=location, paired=True)

        return {
            "reference": reference_receptors,
            "max_edit_distances": max_edit_distances,
            "reads": ReadsType[reads.upper()],
            "sum_matches": sum_matches,
            "normalize": normalize,
            "name": name
        }

    @staticmethod
    def build_object(dataset=None, **params):
        if isinstance(dataset, RepertoireDataset):
            prepared_parameters = MatchedReceptorsEncoder._prepare_parameters(**params)
            return MatchedReceptorsEncoder(**prepared_parameters)
        else:
            raise ValueError(
                "MatchedReceptorsEncoder is not defined for dataset types which are not RepertoireDataset.")

    def encode(self, dataset, params: EncoderParams):
        cache_key = CacheHandler.generate_cache_key(self._prepare_caching_params(dataset, params))
        encoded_dataset = CacheHandler.memo(cache_key,
                                            lambda: self._encode_new_dataset(dataset, params))

        return encoded_dataset

    def _prepare_caching_params(self, dataset, params: EncoderParams):

        chains = [(receptor.chain_1, receptor.chain_2)
                  for receptor in self.reference_receptors]

        encoding_params_desc = {"max_edit_distance": sorted(self.max_edit_distances.items()),
                                "reference_receptors": sorted([getattr(chain_a, params.sequence_type.value)
                                                               + chain_a.v_call +
                                                               chain_a.j_call + "|"
                                                               + getattr(chain_b, params.sequence_type.value) +
                                                               chain_b.v_call + chain_b.j_call
                                                               for chain_a, chain_b in chains]),
                                "reads": self.reads.name,
                                "sum_matches": self.sum_matches,
                                "normalize": self.normalize}

        return (("dataset_identifiers", tuple(dataset.get_example_ids())),
                ("dataset_metadata", dataset.metadata_file),
                ("dataset_type", dataset.__class__.__name__),
                ("labels", tuple(params.label_config.get_labels_by_name())),
                ("encoding", MatchedReceptorsEncoder.__name__),
                ("learn_model", params.learn_model),
                ("encoding_params", encoding_params_desc),)

    def _encode_new_dataset(self, dataset, params: EncoderParams):
        feature_annotations = None if self.sum_matches else self._get_feature_info()

        if self.sum_matches:
            chains = self.reference_receptors[0].chain_pair.value
            feature_names = [f"sum_of_{self.reads.value}_reads_{chains[0]}",
                             f"sum_of_{self.reads.value}_reads_{chains[1]}"]
        else:
            feature_names = [f"{row['cell_id']}.{row['locus']}" for index, row in feature_annotations.iterrows()]

        encoded_repertoires, labels, example_ids = self._encode_repertoires(dataset, params)
        encoded_repertoires = self._normalize(dataset, encoded_repertoires) if self.normalize else encoded_repertoires

        encoded_dataset = dataset.clone()
        encoded_dataset.encoded_data = EncodedData(
            examples=encoded_repertoires,
            example_ids=example_ids,
            feature_names=feature_names,
            feature_annotations=feature_annotations,
            labels=labels,
            encoding=MatchedReceptorsEncoder.__name__,
            info={'sequence_type': params.sequence_type,
                  'region_type': params.region_type}
        )

        return encoded_dataset

    def _normalize(self, dataset, encoded_repertoires):
        if self.reads == ReadsType.UNIQUE:
            repertoire_totals = np.asarray([[repertoire.get_element_count() for repertoire in dataset.get_data()]]).T
        else:
            repertoire_totals = np.asarray(
                [[sum(repertoire.data.duplicate_count) for repertoire in dataset.get_data()]]).T

        return encoded_repertoires / repertoire_totals

    def _get_feature_info(self):
        """
        returns a pandas dataframe containing:
         - receptor id
         - receptor chain
         - amino acid sequence
         - v gene
         - j gene
         - cell id
        """

        features = [[] for i in range(0, self.feature_count)]

        for i, receptor in enumerate(self.reference_receptors):
            id = receptor.receptor_id
            chain_names = receptor.chain_pair.value
            first_chain = receptor.chain_1
            second_chain = receptor.chain_2
            cell_id = receptor.cell_id

            clonotype_id = receptor.metadata["clonotype_id"] if "clonotype_id" in receptor.metadata else None

            if first_chain.metadata is not None:
                first_dual_chain_id = first_chain.metadata[
                    "dual_chain_id"] if "dual_chain_id" in first_chain.metadata else None

            if second_chain.metadata is not None:
                second_dual_chain_id = second_chain.metadata[
                    "dual_chain_id"] if "dual_chain_id" in second_chain.metadata else None

            features[i * 2] = [id, clonotype_id, chain_names[0],
                               first_dual_chain_id,
                               first_chain.sequence_aa,
                               first_chain.v_call,
                               first_chain.j_call, cell_id]
            features[i * 2 + 1] = [id, clonotype_id, chain_names[1],
                                   second_dual_chain_id,
                                   second_chain.sequence_aa,
                                   second_chain.v_call,
                                   second_chain.j_call, cell_id]

        features = pd.DataFrame(features,
                                columns=["receptor_id", "clonotype_id", "locus", "dual_chain_id", "sequence",
                                         "v_call", "j_call", 'cell_id'])

        features.dropna(axis="columns", how="all", inplace=True)

        return features

    def _encode_repertoires(self, dataset: RepertoireDataset, params: EncoderParams):
        # Rows = repertoires, Columns = reference chains (two per sequence receptor)
        encoded_repertories = np.zeros((dataset.get_example_count(),
                                        self.feature_count),
                                       dtype=int)
        labels = {label: [] for label in params.label_config.get_labels_by_name()} if params.encode_labels else None

        for i, repertoire in enumerate(dataset.get_data()):
            encoded_repertories[i] = self._compute_matches_to_reference(repertoire, params)

            if labels is not None:
                for label_name in params.label_config.get_labels_by_name():
                    labels[label_name].append(repertoire.metadata[label_name])

        return encoded_repertories, labels, dataset.get_repertoire_ids()

    def _get_repertoire_matches_to_reference(self, repertoire):
        return CacheHandler.memo_by_params(
            (("repertoire_identifier", repertoire.identifier),
             ("encoding", MatchedReceptorsEncoder.__name__),
             ("readstype", self.reads.name),
             ("sum_matches", self.sum_matches),
             ("max_edit_distances", tuple(self.max_edit_distances.items())),
             ("reference_receptors",
              tuple([self.get_receptor_params(receptor) for receptor in self.reference_receptors]))),
            lambda: self._compute_matches_to_reference(repertoire))

    def get_receptor_params(self, receptor):
        params = []

        for chain in receptor.get_chains():
            receptor_sequence = receptor.get_chain(chain)
            params.append((chain, receptor_sequence.get_sequence(), receptor_sequence.get_attribute("v_gene"),
                           receptor_sequence.get_attribute("j_gene")))

        return tuple(params)

    def _compute_matches_to_reference(self, repertoire: Repertoire, params: EncoderParams):
        matcher = SequenceMatcher()
        matches = np.zeros(self.feature_count, dtype=int)
        rep_seqs = repertoire.sequences(params.region_type)

        for i, ref_receptor in enumerate(self.reference_receptors):
            chain_names = ref_receptor.chain_pair.value
            first_chain = ref_receptor.chain_1
            second_chain = ref_receptor.chain_2

            for rep_seq in rep_seqs:
                matches_idx = 0 if self.sum_matches else i * 2
                match_count = 1 if self.reads == ReadsType.UNIQUE else rep_seq.duplicate_count

                # Match with first chain: add to even columns in matches.
                # Match with second chain: add to odd columns
                if matcher.matches_sequence(first_chain, rep_seq, max_distance=self.max_edit_distances[chain_names[0]]):
                    matches[matches_idx] += match_count
                if matcher.matches_sequence(second_chain, rep_seq,
                                            max_distance=self.max_edit_distances[chain_names[1]]):
                    matches[matches_idx + 1] += match_count

        return matches



import copy
import os
from pathlib import Path
from typing import List

import pandas as pd

from immuneML.data_model.SequenceParams import RegionType
from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.environment.LabelConfiguration import LabelConfiguration
from immuneML.environment.SequenceType import SequenceType
from immuneML.hyperparameter_optimization.config.SplitConfig import SplitConfig
from immuneML.hyperparameter_optimization.core.HPUtil import HPUtil
from immuneML.reports.Report import Report
from immuneML.util.Logger import print_log, log_memory_usage
from immuneML.util.PathBuilder import PathBuilder
from immuneML.workflows.instructions.Instruction import Instruction
from immuneML.workflows.instructions.clustering.ClusteringReportHandler import ClusteringReportHandler
from immuneML.workflows.instructions.clustering.ClusteringRunner import ClusteringRunner
from immuneML.workflows.instructions.clustering.ClusteringState import ClusteringConfig, ClusteringState, \
    ClusteringResultPerRun
from immuneML.workflows.instructions.clustering.ValidationHandler import ValidationHandler
from immuneML.workflows.instructions.clustering.clustering_run_model import ClusteringSetting


class ClusteringInstruction(Instruction):
    """

    Clustering instruction fits clustering methods to the provided encoded dataset and compares the combinations of
    clustering method with its hyperparameters, and encodings across a pre-defined set of metrics. The dataset is split
    into discovery and validation datasets and the clustering results are reported on both. Finally, it
    provides options to include a set of reports to visualize the results.

    See also: :ref:`How to perform clustering analysis`

    For more details on choosing the clustering algorithm and its hyperparameters, see the paper:
    Ullmann, T., Hennig, C., & Boulesteix, A.-L. (2022). Validation of cluster analysis results on validation
    data: A systematic framework. WIREs Data Mining and Knowledge Discovery, 12(3), e1444.
    https://doi.org/10.1002/widm.1444


    **Specification arguments:**

    - dataset (str): name of the dataset to be clustered

    - metrics (list): a list of metrics to use for comparison of clustering algorithms and encodings (it can include
      metrics for either internal evaluation if no labels are provided or metrics for external evaluation so that the
      clusters can be compared against a list of predefined labels); some of the supported metrics include adjusted_rand_score,
      completeness_score, homogeneity_score, silhouette_score; for the full list, see scikit-learn's documentation of
      clustering metrics at https://scikit-learn.org/stable/api/sklearn.metrics.html#module-sklearn.metrics.cluster.

    - labels (list): an optional list of labels to use for external evaluation of clustering

    - split_config (SplitConfig): how to perform splitting of the original dataset into discovery and validation data;
      for this parameter, specify: split_strategy (leave_one_out_stratification, manual, random), training percentage
      if split_strategy is random, and defaults of manual or leave one out stratification config for corresponding split
      strategy; all three options are illustrated here:

      .. indent with spaces
      .. code-block:: yaml

        split_config:
            split_strategy: manual
            manual_config:
                discovery_data: file_with_ids_of_examples_for_discovery_data.csv
                validation_data: file_with_ids_of_examples_for_validation_data.csv

      .. indent with spaces
      .. code-block:: yaml

        split_config:
            split_strategy: random
            training_percentage: 0.5
            split_count: 3 # repeat the random split 3 times -> 3 discovery and 3 validation datasets

      .. indent with spaces
      .. code-block:: yaml

        split_config:
            split_strategy: leave_one_out_stratification
            leave_one_out_config:
                parameter: subject_id # any name of the parameter for split, must be present in the metadata
                min_count: 1 #  defines the minimum number of examples that can be present in the validation dataset.

    - clustering_settings (list): a list where each element represents a :py:obj:`~immuneML.workflows.clustering.clustering_run_model.ClusteringSetting`; a combinations of encoding,
      optional dimensionality reduction algorithm, and the clustering algorithm that will be evaluated

    - reports (list): a list of reports to be run on the clustering results or the encoded data

    - number_of_processes (int): how many processes to use for parallelization

    - sequence_type (str): whether to do analysis on the amino_acid or nucleotide level; this value is used only if
      nothing is specified on the encoder level

    - region_type (str): which part of the receptor sequence to analyze (e.g., IMGT_CDR3); this value is used only if
      nothing is specified on the encoder level

    - validation_type (list): a list of validation types to use for comparison of clustering algorithms and encodings;
      it can be method_based and/or result_based

    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        instructions:
            my_clustering_instruction:
                type: Clustering
                dataset: d1
                metrics: [adjusted_rand_score, adjusted_mutual_info_score]
                labels: [epitope, v_call]
                sequence_type: amino_acid
                region_type: imgt_cdr3
                validation_type: [method_based, result_based]
                split_config:
                    split_count: 1
                    split_strategy: manual
                    manual_config:
                        discovery_data: file_with_ids_of_examples_for_discovery_data.csv
                        validation_data: file_with_ids_of_examples_for_validation_data.csv
                clustering_settings:
                    - encoding: e1
                      dim_reduction: pca
                      method: k_means1
                    - encoding: e2
                      method: dbscan
                reports: [rep1, rep2]

    """

    def __init__(self, dataset: Dataset, metrics: List[str], clustering_settings: List[ClusteringSetting],
                 name: str, label_config: LabelConfiguration = None, reports: List[Report] = None,
                 number_of_processes: int = None, split_config: SplitConfig = None, sequence_type: SequenceType = None,
                 region_type: RegionType = None, validation_type: List[str] = None):

        config = ClusteringConfig(name=name, dataset=dataset, metrics=metrics, clustering_settings=clustering_settings,
                                  label_config=label_config, split_config=split_config, sequence_type=sequence_type,
                                  region_type=region_type, validation_type=validation_type)
        self.number_of_processes = number_of_processes
        self.state = ClusteringState(config=config, name=name)
        self.report_handler = ClusteringReportHandler(reports)
        self.cl_runner = ClusteringRunner(self.state.config, self.number_of_processes, self.report_handler)
        self.validation_handler = ValidationHandler(self.state.config, self.cl_runner, self.report_handler,
                                                    self.number_of_processes)

    def run(self, result_path: Path):
        """Execute the clustering instruction workflow."""
        self._fix_max_processes()
        self._setup_paths(result_path)
        self._split_dataset()

        for run_id in range(self.state.config.split_config.split_count):

            print_log(f"Running clustering for split {run_id + 1}.")

            path = self.state.result_path / f"split_{run_id + 1}"

            # Run discovery
            self._run_discovery(run_id, path / 'discovery')

            log_memory_usage(f"discovery in split {run_id + 1}", f"Clustering instruction {self.state.name}")

            # Run validations

            if self.state.config.split_config.is_with_test_set():

                print_log(f"Running validation for split {run_id + 1}.")

                predictions_df = self._init_predictions_df(self.state.validation_datasets[run_id])

                if "method_based" in self.state.config.validation_type:
                    self.state = self.validation_handler.run_method_based_validation(
                        self.state.validation_datasets[run_id], run_id, PathBuilder.build(path / 'method_based_validation'),
                        copy.deepcopy(predictions_df), self.state)

                    log_memory_usage(f"method-based validation in split {run_id + 1}",
                                     f"Clustering instruction {self.state.name}")

                if "result_based" in self.state.config.validation_type:
                    self.state = self.validation_handler.run_result_based_validation(
                        self.state.validation_datasets[run_id], run_id, PathBuilder.build(path / "result_based_validation"),
                        copy.deepcopy(predictions_df), self.state)

                    log_memory_usage(f"result-based validation in split {run_id + 1}",
                                     f"Clustering instruction {self.state.name}")

                print_log(f"Clustering for split {run_id + 1} finished.")

        self.report_handler.run_clustering_reports(self.state)
        return self.state

    def _run_discovery(self, run_id: int, path: Path):
        """Run clustering on discovery data."""

        print_log("Running clustering on discovery data.")

        dataset = self.state.discovery_datasets[run_id]
        analysis_desc = 'discovery'

        predictions_df = self._init_predictions_df(dataset)
        clustering_items, predictions_df = self.cl_runner.run_all_settings(dataset, analysis_desc, path, run_id,
                                                                           predictions_df, self.state)

        predictions_df.to_csv(self.state.predictions_paths[run_id][analysis_desc], index=False)
        cl_result = ClusteringResultPerRun(run_id, analysis_desc, clustering_items)
        self.state.add_cl_result_per_run(run_id, analysis_desc, cl_result)

        print_log("Clustering on discovery data finished.")

    def _split_dataset(self):
        self.state.discovery_datasets, self.state.validation_datasets = HPUtil.split_data(
            dataset=self.state.config.dataset,
            split_config=self.state.config.split_config,
            path=self.state.result_path,
            label_config=self.state.config.label_config)

    def _setup_paths(self, result_path: Path):
        self.state.result_path = PathBuilder.build(result_path / self.state.config.name)
        self.state.predictions_paths = [
            {'discovery': PathBuilder.build(
                self.state.result_path / f"split_{run_id + 1}") / 'predictions_discovery.csv',
             'result_based_validation': self.state.result_path / f'split_{run_id + 1}/predictions_result_based_validation.csv',
             'method_based_validation': self.state.result_path / f'split_{run_id + 1}/predictions_method_based_validation.csv'}
            for run_id in range(self.state.config.split_config.split_count)]

    def _init_predictions_df(self, dataset: Dataset) -> pd.DataFrame:

        if len(self.state.config.label_config.get_labels_by_name()) > 0:
            predictions_df = dataset.get_metadata(self.state.config.label_config.get_labels_by_name(), return_df=True)
        else:
            predictions_df = pd.DataFrame(index=range(dataset.get_example_count()))

        predictions_df['example_id'] = dataset.get_example_ids()

        return predictions_df

    def _fix_max_processes(self):

        if self.number_of_processes:

            try:
                import torch
                torch.set_num_threads(self.number_of_processes)
            except ImportError as e:
                pass

            os.environ["OMP_NUM_THREADS"] = str(self.number_of_processes)
            os.environ["OPENBLAS_NUM_THREADS"] = str(self.number_of_processes)
            os.environ["MKL_NUM_THREADS"] = str(self.number_of_processes)



import pandas as pd

from immuneML.IO.dataset_import.DataImport import DataImport
from immuneML.data_model.SequenceParams import RegionType
from scripts.specification_util import update_docs_per_mapping


class MiXCRImport(DataImport):
    """
    Imports data in MiXCR format into a Repertoire-, or SequenceDataset.
    RepertoireDatasets should be used when making predictions per repertoire, such as predicting a disease state.
    SequenceDatasets should be used when predicting values for unpaired (single-chain) immune receptors, like
    antigen specificity.


    **Specification arguments:**

    - path (str): For RepertoireDatasets, this is the path to a directory with MiXCR files to import. For Sequence- or ReceptorDatasets this path may either be the path to the file to import, or the path to the folder locating one or multiple files with .tsv, .csv or .txt extensions. By default path is set to the current working directory.

    - is_repertoire (bool): If True, this imports a RepertoireDataset. If False, it imports a SequenceDataset. By default, is_repertoire is set to True.

    - metadata_file (str): Required for RepertoireDatasets. This parameter specifies the path to the metadata file. This is a csv file with columns filename, subject_id and arbitrary other columns which can be used as labels in instructions. Only the MiXCR files included under the column 'filename' are imported into the RepertoireDataset. For setting Sequence- or ReceptorDataset labels, metadata_file is ignored, use label_columns instead.

    - label_columns (list): For Sequence- or ReceptorDataset, this parameter can be used to explicitly set the column names of labels to import. These labels can be used as prediction target. When label_columns are not set, label names are attempted to be discovered automatically (any column name which is not used in the column_mapping). For setting RepertoireDataset labels, label_columns is ignored, use metadata_file instead.

    - import_illegal_characters (bool): Whether to import sequences that contain illegal characters, i.e., characters that do not appear in the sequence alphabet (amino acids including stop codon '*', or nucleotides). When set to false, filtering is only applied to the sequence type of interest (when running immuneML in amino acid mode, only entries with illegal characters in the amino acid sequence, such as '_', are removed). By default import_illegal_characters is False.

    - import_empty_nt_sequences (bool): imports sequences which have an empty nucleotide sequence field; can be True or False. By default, import_empty_nt_sequences is set to True.

    - import_empty_aa_sequences (bool): imports sequences which have an empty amino acid sequence field; can be True or False; for analysis on amino acid sequences, this parameter should be False (import only non-empty amino acid sequences). By default, import_empty_aa_sequences is set to False.

    - region_type (str): Which part of the sequence to import. By default, this value is set to IMGT_CDR3. This means the first and last amino acids are removed from the CDR3 sequence, as MiXCR format contains the trailing and leading conserved amino acids in the CDR3. Valid values for region_type are the names of the :py:obj:`~immuneML.data_model.receptor.RegionType.RegionType` enum.

    - column_mapping (dict): A mapping from MiXCR column names to immuneML's internal data representation. The columns that specify the sequences to import are handled by the region_type parameter. A custom column mapping can be specified here if necessary (for example; adding additional data fields if they are present in the MiXCR file, or using alternative column names). Valid immuneML fields that can be specified here are defined by the AIRR standard (AIRRSequenceSet). For MiXCR, this is by default set to:

        .. indent with spaces
        .. code-block:: yaml

            cloneCount: duplicate_count
            allVHitsWithScore: v_call
            allJHitsWithScore: j_call
            aaSeqCDR3: junction_aa
            nSeqCDR3: junction
            aaSeqCDR1: cdr1_aa
            nSeqCDR1: cdr1
            aaSeqCDR2: cdr2_aa
            nSeqCDR2: cdr2

    - columns_to_load (list): Specifies which subset of columns must be loaded from the MiXCR file. By default, this is: [cloneCount, allVHitsWithScore, allJHitsWithScore, aaSeqCDR3, nSeqCDR3]

    - separator (str): Column separator, for MiXCR this is by default "\\t".


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            datasets:
                my_mixcr_dataset:
                    format: MiXCR
                    params:
                        path: path/to/files/
                        is_repertoire: True # whether to import a RepertoireDataset (True) or a SequenceDataset (False)
                        metadata_file: path/to/metadata.csv # metadata file for RepertoireDataset
                        region_type: IMGT_CDR3 # what part of the sequence to import
                        import_illegal_characters: False # remove sequences with illegal characters for the sequence_type being used
                        import_empty_nt_sequences: True # keep sequences even though the nucleotide sequence might be empty
                        import_empty_aa_sequences: False # filter out sequences if they don't have sequence_aa set
                        # Optional fields with MiXCR-specific defaults, only change when different behavior is required:
                        separator: "\\t" # column separator
                        columns_to_load: # subset of columns to load, sequence columns are handled by region_type parameter
                        - cloneCount
                        - allVHitsWithScore
                        - allJHitsWithScore
                        - aaSeqCDR3
                        - nSeqCDR3
                        column_mapping: # column mapping MiXCR: immuneML
                            cloneCount: duplicate_count
                            allVHitsWithScore: v_call
                            allJHitsWithScore: j_call
                            mixcrColumnName1: metadata_label1
                            mixcrColumnName2: metadata_label2

    """

    def preprocess_file(self, df: pd.DataFrame) -> pd.DataFrame:
        df["v_call"] = load_alleles(df, "v_call")
        df["j_call"] = load_alleles(df, "j_call")

        return df

    @staticmethod
    def get_documentation():
        doc = str(MiXCRImport.__doc__)

        region_type_values = str([region_type.name for region_type in RegionType])[1:-1].replace("'", "`")

        mapping = {
            "Valid values for region_type are the names of the :py:obj:`~immuneML.data_model.receptor.RegionType.RegionType` enum.": f"Valid values are {region_type_values}.",
            "Valid immuneML fields that can be specified here are defined by the AIRR standard (AIRRSequenceSet)": f"Valid immuneML fields that can be specified here by `the AIRR Rearrangement Schema <https://docs.airr-community.org/en/latest/datarep/rearrangements.html>`_."
        }
        doc = update_docs_per_mapping(doc, mapping)
        return doc



def load_alleles(df: pd.DataFrame, column_name):
    # note: MiXCR omits the '/' for 'TRA.../DV' genes, and remove "*00" for allele if set
    tmp_df = df.apply(
        lambda row: row[column_name].split(",")[0].split("(")[0].replace("DV", "/DV").replace("//",
                                                                                              "/").replace(
            r'*00', ''), axis=1)
    return tmp_df



import copy
import logging
from pathlib import Path
from typing import Tuple, Dict, Union

import numpy as np
import pandas as pd
import sklearn
from scipy.sparse import issparse

from immuneML.caching.CacheHandler import CacheHandler
from immuneML.data_model.SequenceParams import RegionType
from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.encodings.DatasetEncoder import DatasetEncoder
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.environment.LabelConfiguration import LabelConfiguration
from immuneML.environment.SequenceType import SequenceType
from immuneML.ml_methods.clustering.ClusteringMethod import ClusteringMethod
from immuneML.ml_metrics.ClusteringMetric import is_external, is_internal
from immuneML.util.Logger import print_log
from immuneML.util.PathBuilder import PathBuilder
from immuneML.workflows.instructions.clustering.ClusteringReportHandler import ClusteringReportHandler
from immuneML.workflows.instructions.clustering.ClusteringState import ClusteringConfig, ClusteringState, \
    ClusteringItemResult
from immuneML.workflows.instructions.clustering.clustering_run_model import ClusteringItem, DataFrameWrapper, \
    ClusteringSetting
from immuneML.workflows.steps.DataEncoder import DataEncoder
from immuneML.workflows.steps.DataEncoderParams import DataEncoderParams


class ClusteringRunner:
    """Handles core clustering operations like fitting, prediction and evaluation."""

    def __init__(self, config: ClusteringConfig, n_processes: int, report_handler: ClusteringReportHandler):
        self.config = config
        self.report_handler = report_handler
        self.number_of_processes = n_processes

    def run_all_settings(self, dataset: Dataset, analysis_desc: str, path: Path, run_id: int,
                         predictions_df: pd.DataFrame, state: ClusteringState):
        clustering_items = {}

        for cl_setting in self.config.clustering_settings:
            cl_item, predictions_df = self.run_setting(dataset, cl_setting, analysis_desc, path, run_id,
                                                       predictions_df, state)
            clustering_items[cl_setting.get_key()] = cl_item

        return clustering_items, predictions_df

    def run_setting(self, dataset: Dataset, cl_setting: ClusteringSetting, analysis_desc: str, path: Path,
                    run_id: int, predictions_df: pd.DataFrame, state: ClusteringState) \
            -> Tuple[ClusteringItemResult, pd.DataFrame]:

        print_log(f"Running clustering setting {cl_setting.get_key()}")

        cl_setting.path = PathBuilder.build(path / f"{cl_setting.get_key()}")

        learn_model = analysis_desc == 'discovery' or analysis_desc == 'method_based_validation'

        # Encode data
        encoder = copy.deepcopy(cl_setting.encoder) if learn_model \
            else state.clustering_items[run_id].discovery.get_cl_item(cl_setting).encoder
        enc_dataset = encode_dataset(dataset, cl_setting, self.number_of_processes, self.config.label_config,
                                     learn_model, self.config.sequence_type, self.config.region_type,
                                     encoder)

        # Run clustering
        method = copy.deepcopy(cl_setting.clustering_method)
        predictions = self._fit_and_predict(enc_dataset, method)
        predictions_df[f'predictions_{cl_setting.get_key()}'] = predictions

        print_log(f"{cl_setting.get_key()}: clustering method fitted and predictions made.")

        # Evaluate results
        features = get_features(enc_dataset, cl_setting)
        performance_paths = self.evaluate_clustering(predictions_df, cl_setting, features)

        cl_item = ClusteringItem(
            cl_setting=cl_setting,
            dataset=enc_dataset,
            predictions=predictions,
            encoder=encoder,
            method=method,
            external_performance=DataFrameWrapper(path=performance_paths['external']) if performance_paths['external'] else None,
            internal_performance=DataFrameWrapper(path=performance_paths['internal']) if performance_paths['internal'] else None
        )

        report_results = self.report_handler.run_item_reports(cl_item, analysis_desc, run_id, cl_setting.path, state)

        print_log(f"Clustering setting {cl_setting.get_key()} finished.")

        return ClusteringItemResult(cl_item, report_results), predictions_df

    def _fit_and_predict(self, dataset: Dataset, method: ClusteringMethod) -> np.ndarray:
        """Fit clustering method and get predictions."""
        if hasattr(method, 'fit_predict'):
            return method.fit_predict(dataset)
        else:
            method.fit(dataset)
            return method.predict(dataset)

    def _eval_internal_metrics(self, predictions: pd.DataFrame, cl_setting: ClusteringSetting, features) -> Path:
        internal_performances = {}
        dense_features = features.toarray() if issparse(features) else features

        for metric in [m for m in self.config.metrics if is_internal(m)]:
            try:
                metric_fn = getattr(sklearn.metrics, metric)
                internal_performances[metric] = [metric_fn(dense_features,
                                                           predictions[f'predictions_{cl_setting.get_key()}'].values)]
            except ValueError as e:
                logging.warning(f"Error calculating metric {metric}: {e}")
                internal_performances[metric] = [np.nan]

        pd.DataFrame(internal_performances).to_csv(str(cl_setting.path / 'internal_performances.csv'), index=False)
        return cl_setting.path / 'internal_performances.csv'

    def _eval_external_metrics(self, predictions: pd.DataFrame, cl_setting: ClusteringSetting) -> Union[Path, None]:
        if self.config.label_config is not None and self.config.label_config.get_label_count() > 0:

            external_performances = {label: {} for label in self.config.label_config.get_labels_by_name()}

            for metric in [m for m in self.config.metrics if is_external(m)]:

                metric_fn = getattr(sklearn.metrics, metric)
                for label in self.config.label_config.get_labels_by_name():
                    try:
                        external_performances[label][metric] = metric_fn(predictions[label].values,
                                                                         predictions[
                                                                             f'predictions_{cl_setting.get_key()}'].values)
                    except ValueError as e:
                        logging.warning(f"Error calculating metric {metric}: {e}")
                        external_performances[label][metric] = np.nan

            (pd.DataFrame(external_performances).reset_index().rename(columns={'index': 'metric'})
             .to_csv(str(cl_setting.path / 'external_performances.csv'), index=False))

            return cl_setting.path / 'external_performances.csv'
        else:
            return None

    def evaluate_clustering(self, predictions: pd.DataFrame, cl_setting: ClusteringSetting, features) \
            -> Dict[str, Path]:

        return {'internal': self._eval_internal_metrics(predictions, cl_setting, features),
                'external': self._eval_external_metrics(predictions, cl_setting)}


def get_features(dataset: Dataset, cl_setting: ClusteringSetting):
    """Get features from encoded dataset."""
    return dataset.encoded_data.examples if cl_setting.dim_reduction_method is None \
        else dataset.encoded_data.dimensionality_reduced_data


def encode_dataset(dataset: Dataset, cl_setting: ClusteringSetting, number_of_processes: int,
                   label_config: LabelConfiguration, learn_model: bool, sequence_type: SequenceType,
                   region_type: RegionType, encoder: DatasetEncoder = None):
    def dict_to_sorted_tuple(d: dict) -> tuple:
        return tuple(sorted(d.items(), key=lambda tup: tup[0])) if d is not None else None

    return CacheHandler.memo_by_params(
        (dataset.identifier, dict_to_sorted_tuple(cl_setting.encoder_params), label_config.get_labels_by_name(),
         sequence_type.name, region_type.name, cl_setting.dim_red_name, learn_model,
         dict_to_sorted_tuple(cl_setting.dim_red_params)),
        lambda: encode_dataset_internal(dataset, cl_setting, number_of_processes, label_config,
                                        learn_model, sequence_type, region_type, encoder))


def encode_dataset_internal(dataset: Dataset, cl_setting: ClusteringSetting, number_of_processes: int,
                            label_config: LabelConfiguration, learn_model: bool, sequence_type: SequenceType,
                            region_type: RegionType, encoder: DatasetEncoder = None):
    enc_params = EncoderParams(model=cl_setting.encoder_params, result_path=cl_setting.path,
                               pool_size=number_of_processes, label_config=label_config,
                               learn_model=learn_model, encode_labels=False,
                               sequence_type=sequence_type,
                               region_type=region_type)
    enc_dataset = DataEncoder.run(DataEncoderParams(dataset=dataset,
                                                    encoder=cl_setting.encoder if encoder is None else encoder,
                                                    encoder_params=enc_params))

    if cl_setting.dim_reduction_method:
        enc_dataset.encoded_data.dimensionality_reduced_data = cl_setting.dim_reduction_method.fit_transform(
            enc_dataset)
        enc_dataset.encoded_data.dim_names = cl_setting.dim_reduction_method.get_dimension_names()

    return enc_dataset



from pathlib import Path

import logging
import pandas as pd
import os

from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.data_model.datasets.ElementDataset import SequenceDataset
from immuneML.dsl.instruction_parsers.LabelHelper import LabelHelper
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.encodings.motif_encoding.MotifEncoder import MotifEncoder
from immuneML.encodings.motif_encoding.PositionalMotifHelper import PositionalMotifHelper
from immuneML.ml_methods.util.Util import Util
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.data_reports.DataReport import DataReport
from immuneML.util.EncoderHelper import EncoderHelper
from immuneML.util.MotifPerformancePlotHelper import MotifPerformancePlotHelper
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder
from immuneML.workflows.steps.DataEncoder import DataEncoder
from immuneML.workflows.steps.DataEncoderParams import DataEncoderParams


class MotifGeneralizationAnalysis(DataReport):
    """
    This report splits the given dataset into a training and validation set, identifies significant motifs using the
    :py:obj:`~immuneML.encodings.motif_encoding.MotifEncoder.MotifEncoder`
    on the training set and plots the precision/recall and precision/true positive predictions of motifs
    on both the training and validation sets. This can be used to:

    - determine the optimal recall cutoff for motifs of a given size
    - investigate how well motifs learned on a training set generalize to a test set

    After running this report and determining the optimal recall cutoffs, the report
    :py:obj:`~immuneML.reports.encoding_reports.MotifTestSetPerformance.MotifTestSetPerformance` can be run to
    plot the performance on an independent test set.

    Note: the MotifEncoder (and thus this report) can only be used for sequences of the same length.

    **Specification arguments:**

    - label (dict): A label configuration. One label should be specified, and the positive_class for this label should be defined. See the YAML specification below for an example.

    - training_set_identifier_path (str): Path to a file containing 'sequence_identifiers' of the sequences used for the training set. This file should have a single column named 'example_id' and have one sequence identifier per line. If training_set_identifier_path is not set, a random subset of the data (according to training_percentage) will be assigned to be the training set.

    - training_percentage (float): If training_set_identifier_path is not set, this value is used to specify the fraction of sequences that will be randomly assigned to form the training set. Should be a value between 0 and 1. By default, training_percentage is 0.7.

    - random_seed (int): Random seed for splitting the data into training and validation sets a training_set_identifier_path is not provided.

    - split_by_motif_size (bool): Whether to split the analysis per motif size. If true, a recall threshold is learned for each motif size, and figures are generated for each motif size independently. By default, split_by_motif_size is true.

    - min_precision: :py:obj:`~immuneML.encodings.motif_encoding.MotifEncoder.MotifEncoder` parameter. The minimum precision threshold for keeping a motif on the training set. By default, min_precision is 0.9.

    - test_precision_threshold (float). The desired precision on the test set, given that motifs are learned by using a training set with a precision threshold of min_precision. It is recommended for test_precision_threshold to be lower than min_precision, e.g., min_precision - 0.1. By default, test_precision_threshold is 0.8.

    - min_recall (float): :py:obj:`~immuneML.encodings.motif_encoding.MotifEncoder.MotifEncoder` parameter. The minimum recall threshold for keeping a motif. Any learned recall threshold will be at least as high as the set min_recall value. The default value for min_recall is 0.

    - min_true_positives (int): :py:obj:`~immuneML.encodings.motif_encoding.MotifEncoder.MotifEncoder` parameter. The minimum number of true positive training sequences that a motif needs to occur in. The default value for min_true_positives is 1.

    - max_positions (int): :py:obj:`~immuneML.encodings.motif_encoding.MotifEncoder.MotifEncoder` parameter. The maximum motif size. This is number of positional amino acids the motif consists of (excluding gaps). The default value for max_positions is 4.

    - min_positions (int): :py:obj:`~immuneML.encodings.motif_encoding.MotifEncoder.MotifEncoder` parameter. The minimum motif size (see also: max_positions). The default value for min_positions is 1.

    - no_gaps (bool): :py:obj:`~immuneML.encodings.motif_encoding.MotifEncoder.MotifEncoder` parameter. Must be set to True if only contiguous motifs (position-specific k-mers) are allowed. By default, no_gaps is False, meaning both gapped and ungapped motifs are searched for.

    - smoothen_combined_precision (bool): whether to add a smoothed line representing the combined precision to the precision-vs-TP plot. When set to True, this may take considerable extra time to compute. By default, plot_smoothed_combined_precision is set to True.

    - min_points_in_window (int): Parameter for smoothing the combined_precision line in the precision-vs-TP plot through lognormal kernel smoothing with adaptive window size. This parameter determines the minimum number of points that need to be present in a window to determine the adaptive window size. By default, min_points_in_window is 50.

    - smoothing_constant1: Parameter for smoothing the combined_precision line in the precision-vs-TP plot through lognormal kernel smoothing with adaptive window size. This smoothing constant determines the dependence of the smoothness on the window size. Increasing this increases smoothness for regions where few points are present. By default, smoothing_constant1 is 5.

    - smoothing_constant2: Parameter for smoothing the combined_precision line in the precision-vs-TP plot through lognormal kernel smoothing. with adaptive window size. This smoothing constant can be used to scale the overall kernel width, thus influencing the smoothness of all regions regardless of data density. By default, smoothing_constant2 is 10.

    - training_set_name (str): Name of the training set to be used in figures. By default, the training_set_name is 'training set'.

    - test_set_name (str): Name of the test set to be used in figures. By default, the test_set_name is 'test set'.

    - highlight_motifs_path (str): Path to a set of motifs of interest to highlight in the output figures (such as implanted ground-truth motifs). By default, no motifs are highlighted.

    - highlight_motifs_name (str): IF highlight_motifs_path is defined, this name will be used to label the motifs of interest in the output figures.


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            reports:
                my_motif_generalization:
                    MotifGeneralizationAnalysis:
                        min_precision: 0.9
                        min_recall: 0.1
                        label: # Define a label, and the positive class for that given label
                            CMV:
                                positive_class: +
    """

    def __init__(self, training_set_identifier_path: str = None, training_percentage: float = None,
                 max_positions: int = None, min_positions: int = None, min_precision: float = None, min_recall: float = None,
                 min_true_positives: int = None, no_gaps: bool = False,
                 test_precision_threshold: float = None,
                 split_by_motif_size: bool = None, random_seed: int = None, label: dict = None,
                 min_points_in_window: int = None, smoothing_constant1: float = None, smoothing_constant2: float = None,
                 highlight_motifs_path: str = None, highlight_motifs_name: str = None,
                 training_set_name: str = None, test_set_name: str = None,
                 dataset: SequenceDataset = None, result_path: Path = None, number_of_processes: int = 1, name: str = None):
        super().__init__(dataset=dataset, result_path=result_path, number_of_processes=number_of_processes, name=name)
        self.training_set_identifier_path = Path(training_set_identifier_path) if training_set_identifier_path is not None else None
        self.training_percentage = training_percentage
        self.max_positions = max_positions
        self.min_positions = min_positions
        self.min_precision = min_precision
        self.test_precision_threshold = test_precision_threshold
        self.min_recall = min_recall
        self.min_true_positives = min_true_positives
        self.no_gaps = no_gaps
        self.split_by_motif_size = split_by_motif_size
        self.min_points_in_window = min_points_in_window
        self.smoothing_constant1 = smoothing_constant1
        self.smoothing_constant2 = smoothing_constant2
        self.random_seed = random_seed
        self.label = label
        self.n_positives_in_training_data = None

        self.training_set_name = training_set_name
        self.test_set_name = test_set_name
        self.highlight_motifs_name = highlight_motifs_name
        self.highlight_motifs_path = Path(highlight_motifs_path) if highlight_motifs_path is not None else None

    @classmethod
    def build_object(cls, **kwargs):
        location = MotifGeneralizationAnalysis.__name__

        ParameterValidator.assert_type_and_value(kwargs["max_positions"], int, location, "max_positions", min_inclusive=1)
        ParameterValidator.assert_type_and_value(kwargs["min_positions"], int, location, "min_positions", min_inclusive=1)
        ParameterValidator.assert_type_and_value(kwargs["no_gaps"], bool, location, "no_gaps")
        assert kwargs["max_positions"] >= kwargs["min_positions"], f"{location}: max_positions ({kwargs['max_positions']}) must be greater than or equal to min_positions ({kwargs['min_positions']})"

        ParameterValidator.assert_type_and_value(kwargs["min_precision"], (int, float), location, "min_precision", min_inclusive=0, max_inclusive=1)
        ParameterValidator.assert_type_and_value(kwargs["min_recall"], (int, float), location, "min_recall", min_inclusive=0, max_inclusive=1)
        ParameterValidator.assert_type_and_value(kwargs["min_true_positives"], int, location, "min_true_positives", min_inclusive=1)
        ParameterValidator.assert_type_and_value(kwargs["test_precision_threshold"], float, location, "test_precision_threshold", min_inclusive=0, max_exclusive=1)
        ParameterValidator.assert_type_and_value(kwargs["split_by_motif_size"], bool, location, "split_by_motif_size")
        ParameterValidator.assert_type_and_value(kwargs["min_points_in_window"], int, location, "min_points_in_window", min_inclusive=1)
        ParameterValidator.assert_type_and_value(kwargs["smoothing_constant1"], (int, float), location, "smoothing_constant1", min_exclusive=0)
        ParameterValidator.assert_type_and_value(kwargs["smoothing_constant2"], (int, float), location, "smoothing_constant2", min_exclusive=0)

        ParameterValidator.assert_type_and_value(kwargs["training_set_name"], str, location, "training_set_name")
        ParameterValidator.assert_type_and_value(kwargs["test_set_name"], str, location, "test_set_name")

        assert kwargs["training_set_name"] != kwargs["test_set_name"], f"{location}: training_set_name cannot be the same as test_set_name. Both are: {kwargs['training_set_name']}"

        if kwargs["training_set_identifier_path"] is not None:
            ParameterValidator.assert_type_and_value(kwargs["training_set_identifier_path"], str, location, "training_set_identifier_path")
            assert os.path.isfile(kwargs["training_set_identifier_path"]), f"{location}: the file {kwargs['training_set_identifier_path']} does not exist. " \
                                         f"Specify the correct path under training_set_identifier_path."
        else:
            ParameterValidator.assert_type_and_value(kwargs["training_percentage"], float, location, "training_percentage", min_exclusive=0, max_exclusive=1)

        if "random_seed" in kwargs and kwargs["random_seed"] is not None:
            ParameterValidator.assert_type_and_value(kwargs["random_seed"], int, location, "random_seed")

        ParameterValidator.assert_type_and_value(kwargs["label"], (dict, str), location, "label")
        if type(kwargs["label"]) is dict:
           assert len(kwargs["label"]) == 1, f"{location}: The number of specified labels must be 1, found {len(kwargs['label'])}: {', '.join(list(len(kwargs['label'].keys())))}"

        if "highlight_motifs_path" in kwargs and kwargs["highlight_motifs_path"] is not None:
            PositionalMotifHelper.check_motif_filepath(kwargs["highlight_motifs_path"], location, "highlight_motifs_path")

        ParameterValidator.assert_type_and_value(kwargs["highlight_motifs_name"], str, location, "highlight_motifs_name")

        return MotifGeneralizationAnalysis(**kwargs)

    def _generate(self):
        encoded_training_dataset, encoded_test_dataset = self._get_encoded_train_test_datasets()
        training_plotting_data, test_plotting_data = MotifPerformancePlotHelper.get_plotting_data(encoded_training_dataset.encoded_data,
                                                                                                  encoded_test_dataset.encoded_data,
                                                                                                  self.highlight_motifs_path, self.highlight_motifs_name)

        self.n_positives_in_training_data = self._get_positive_count(encoded_training_dataset)

        return self._get_report_result(training_plotting_data, test_plotting_data)

    def _get_report_result(self, training_plotting_data, test_plotting_data):
        if self.split_by_motif_size:
            output_tables, output_plots, tp_cutoff_dict = self._construct_and_plot_data_per_motif_size(training_plotting_data, test_plotting_data)
        else:
            output_tables, output_plots, tp_cutoff_dict = self._construct_and_plot_data(training_plotting_data, test_plotting_data)

        tp_cutoff_table = self._write_tp_recall_thresholds(tp_cutoff_dict)
        output_tables.append(tp_cutoff_table)

        return ReportResult(output_tables=output_tables,
                            output_figures=output_plots)

    def _construct_and_plot_data_per_motif_size(self, training_plotting_data, test_plotting_data):
        output_tables, output_plots = [], []
        tp_cutoff_dict = {}

        training_plotting_data["motif_size"] = training_plotting_data["feature_names"].apply(PositionalMotifHelper.get_motif_size)
        test_plotting_data["motif_size"] = test_plotting_data["feature_names"].apply(PositionalMotifHelper.get_motif_size)

        for motif_size in sorted(set(training_plotting_data["motif_size"])):
            sub_training_plotting_data = training_plotting_data[training_plotting_data["motif_size"] == motif_size]
            sub_test_plotting_data = test_plotting_data[test_plotting_data["motif_size"] == motif_size]

            sub_output_tables, sub_output_plots, sub_tp_cutoff_dict = self._construct_and_plot_data(sub_training_plotting_data, sub_test_plotting_data, motif_size=motif_size)

            output_tables.extend(sub_output_tables)
            output_plots.extend(sub_output_plots)
            tp_cutoff_dict.update(sub_tp_cutoff_dict)

        return output_tables, output_plots, tp_cutoff_dict


    def _construct_and_plot_data(self, training_plotting_data, test_plotting_data, motif_size=None):
        training_combined_precision = self._get_combined_precision(training_plotting_data)
        test_combined_precision = self._get_combined_precision(test_plotting_data)
        tp_cutoff = self._determine_tp_cutoff(test_combined_precision, motif_size)

        motif_size_suffix = f"_motif_size={motif_size}" if motif_size is not None else ""
        motifs_name = f"motifs of length {motif_size}" if motif_size is not None else "motifs"

        output_tables = MotifPerformancePlotHelper.write_output_tables(self, training_plotting_data, test_plotting_data, training_combined_precision, test_combined_precision, motifs_name=motifs_name, file_suffix=motif_size_suffix)
        output_plots = MotifPerformancePlotHelper.write_plots(self, training_plotting_data, test_plotting_data, training_combined_precision, test_combined_precision, training_tp_cutoff=None, test_tp_cutoff=tp_cutoff, motifs_name=motifs_name, file_suffix=motif_size_suffix)

        return output_tables, output_plots, {motif_size: tp_cutoff}

    def _write_tp_recall_thresholds(self, tp_cutoff_dict):
        training_set_name = self.training_set_name.replace(" ", "_")

        data = {"precision_cutoff": [self.min_precision] * len(tp_cutoff_dict),
                "n_positives_in_training_data": [self.n_positives_in_training_data] * len(tp_cutoff_dict),
                f"{training_set_name}_tp_cutoff": [],
                f"{training_set_name}_recall_cutoff": [],
                "motif_size": []}

        for motif_size, tp_cutoff in tp_cutoff_dict.items():
            data["motif_size"].append(motif_size if motif_size is not None else "all")
            data[f"{training_set_name}_tp_cutoff"].append(tp_cutoff)
            data[f"{training_set_name}_recall_cutoff"].append(self._tp_to_recall(tp_cutoff))

        return self._write_output_table(table=pd.DataFrame(data),
                                        file_path=self.result_path / "tp_recall_cutoffs.tsv",
                                        name=f"{self.training_set_name}-TP and recall cutoff(s)")
    def _tp_to_recall(self, tp_cutoff):
        if tp_cutoff is not None:
            return tp_cutoff / self.n_positives_in_training_data

    def _get_encoded_train_test_datasets(self):
        train_data_path = PathBuilder.build(self.result_path / "datasets/train")
        test_data_path = PathBuilder.build(self.result_path / "datasets/test")

        train_indices, val_indices = self._get_train_val_indices()

        training_data = self.dataset.make_subset(train_indices, train_data_path, Dataset.TRAIN)
        test_data = self.dataset.make_subset(val_indices, test_data_path, Dataset.TEST)

        encoder = self._get_encoder()

        encoded_training_dataset = self._encode_dataset(training_data, encoder, learn_model=True)
        encoded_test_dataset = self._encode_dataset(test_data, encoder, learn_model=False)

        return encoded_training_dataset, encoded_test_dataset

    def _get_train_val_indices(self):
        if self.training_set_identifier_path is None:
            return Util.get_train_val_indices(self.dataset.get_example_count(),
                                              self.training_percentage, random_seed=self.random_seed)
        else:
            return self._get_train_val_indices_from_file()


    def _get_train_val_indices_from_file(self):
        input_train_identifiers = list(pd.read_csv(self.training_set_identifier_path, usecols=["example_id"])["example_id"].astype(str))

        train_indices = []
        val_indices = []
        val_identifiers = []
        actual_train_identifiers = []

        for idx, sequence in enumerate(self.dataset.get_data()):
            if str(sequence.sequence_id) in input_train_identifiers:
                train_indices.append(idx)
                actual_train_identifiers.append(sequence.sequence_id)
            else:
                val_indices.append(idx)
                val_identifiers.append(sequence.sequence_id)

        self._write_identifiers(self.result_path / "training_set_identifiers.txt", actual_train_identifiers, "Training")
        self._write_identifiers(self.result_path / "validation_set_identifiers.txt", val_identifiers, "Validation")

        assert len(train_indices) > 0, f"{MotifGeneralizationAnalysis.__name__}: error when reading training set identifiers from training_set_identifier_path, 0 of the identifiers were present in the dataset. Please check training_set_identifier_path: {self.training_set_identifier_path}, and see the log file for more information."
        assert len(val_indices) > 0, f"{MotifGeneralizationAnalysis.__name__}: error when inferring validation set identifiers from training_set_identifier_path, all of the identifiers were present in the dataset resulting in 0 sequences in the validation set. Please check training_set_identifier_path: {self.training_set_identifier_path}, and see the log file for more information."
        assert len(train_indices) == len(input_train_identifiers), f"{MotifGeneralizationAnalysis.__name__}: error when reading training set identifiers from training_set_identifier_path, not all identifiers provided in the file occurred in the dataset ({len(train_indices)} of {len(input_train_identifiers)} found). Please check training_set_identifier_path: {self.training_set_identifier_path}, and see the log file for more information."

        return train_indices, val_indices

    def _write_identifiers(self, path, identifiers, set_name):
        logging.info(f"{MotifGeneralizationAnalysis.__name__}: {len(identifiers)} {set_name} set identifiers written to: {path}")

        with open(path, "w") as file:
            file.writelines([f"{identifier}\n" for identifier in identifiers])

    def _get_encoder(self):
        encoder = MotifEncoder.build_object(self.dataset, **{"max_positions": self.max_positions,
                                                             "min_positions": self.min_positions,
                                                            "min_precision": self.min_precision,
                                                            "min_recall": self.min_recall,
                                                            "min_true_positives": self.min_true_positives,
                                                             "no_gaps": self.no_gaps,
                                                            "label": None,
                                                            "name": f"motif_encoder"})

        return encoder

    def _encode_dataset(self, dataset, encoder, learn_model):
        encoded_dataset = DataEncoder.run(DataEncoderParams(dataset=dataset, encoder=encoder,
                                                            encoder_params=EncoderParams(result_path=self.result_path / f"encoded_data/{dataset.name}",
                                                                                         label_config=self._get_label_config(dataset),
                                                                                         pool_size=self.number_of_processes,
                                                                                         learn_model=learn_model,
                                                                                         encode_labels=True),
                                                            ))

        return encoded_dataset

    def _get_label_config(self, dataset):
        label_config = LabelHelper.create_label_config([self.label], dataset, MotifGeneralizationAnalysis.__name__,
                                                       f"{MotifGeneralizationAnalysis.__name__}/label")
        EncoderHelper.check_positive_class_labels(label_config, f"{MotifGeneralizationAnalysis.__name__}/label")

        return label_config

    def _get_positive_count(self, dataset):
        label_config = self._get_label_config(dataset)
        label_name = label_config.get_label_objects()[0].name
        label_positive_class = label_config.get_label_objects()[0].positive_class

        return sum([1 for label_class in dataset.get_metadata([label_name])[label_name] if label_class == label_positive_class])

    def _get_combined_precision(self, plotting_data):
        return MotifPerformancePlotHelper.get_combined_precision(plotting_data,
                                                                 min_points_in_window=self.min_points_in_window,
                                                                 smoothing_constant1=self.smoothing_constant1,
                                                                 smoothing_constant2=self.smoothing_constant2)

    def _determine_tp_cutoff(self, combined_precision, motif_size=None):
        col = "smooth_combined_precision" if "smooth_combined_precision" in combined_precision else "combined_precision"

        try:
            # assert all(training_combined_precision["training_TP"] == test_combined_precision["training_TP"])
            #
            # train_test_difference = training_combined_precision[col] - test_combined_precision[col]
            # return min(test_combined_precision[train_test_difference <= self.precision_difference]["training_TP"])

            max_tp_below_threshold = max(combined_precision[combined_precision[col] < self.test_precision_threshold]["training_TP"])
            all_above_threshold = combined_precision[combined_precision["training_TP"] > max_tp_below_threshold]

            return min(all_above_threshold["training_TP"])
        except ValueError:
            motif_size_warning = f" for motif size = {motif_size}" if motif_size is not None else ""
            logging.warning(f"{MotifGeneralizationAnalysis.__name__}: could not automatically determine optimal TP threshold{motif_size_warning} with precison differenc  based on {col}")
            return None

    def _plot_precision_per_tp(self, file_path, plotting_data, combined_precision, dataset_type, tp_cutoff, motifs_name="motifs"):
         return MotifPerformancePlotHelper.plot_precision_per_tp(file_path, plotting_data, combined_precision, dataset_type,
                                                                 training_set_name=self.training_set_name,
                                                                 tp_cutoff=tp_cutoff, motifs_name=motifs_name,
                                                                 highlight_motifs_name=self.highlight_motifs_name)

    def _plot_precision_recall(self, file_path, plotting_data, min_recall=None, min_precision=None, dataset_type=None, motifs_name="motifs"):
        return MotifPerformancePlotHelper.plot_precision_recall(file_path, plotting_data, min_recall=min_recall, min_precision=min_precision,
                                                                 dataset_type=dataset_type, motifs_name=motifs_name, highlight_motifs_name=self.highlight_motifs_name)



import logging
import inspect

from immuneML.ml_metrics import ml_metrics
from sklearn import metrics as sklearn_metrics
from immuneML.ml_metrics.ClassificationMetric import ClassificationMetric
from immuneML.ml_methods.util.Util import Util
from immuneML.environment.Constants import Constants


class MetricUtil:

    @staticmethod
    def get_metric_fn(metric: ClassificationMetric):
        if hasattr(ml_metrics, metric.value):
            fn = getattr(ml_metrics, metric.value)
        else:
            fn = getattr(sklearn_metrics, metric.value)

        return fn

    @staticmethod
    def score_for_metric(metric: ClassificationMetric, predicted_y, predicted_proba_y, true_y, classes, pos_class=None):
        """
        Note: when providing label classes, make sure the 'positive class' is sorted last.
        This sorting should be done automatically when accessing Label.values
        """

        fn = MetricUtil.get_metric_fn(metric)

        if metric in ClassificationMetric.get_binary_only_metrics():
            true_y, predicted_y = Util.binarize_label_classes(true_y=true_y, predicted_y=predicted_y, classes=classes)

        try:
            if metric in ClassificationMetric.get_probability_based_metric_types():
                predictions = predicted_proba_y
                if predicted_proba_y is None:
                    logging.warning(
                        f"MLMethodAssessment: metric {metric} is specified, but the chosen ML method does not output "
                        f"class probabilities. Using predicted classes instead...")
                    predictions = predicted_y
            else:
                predictions = predicted_y

            if 'labels' in inspect.getfullargspec(fn).kwonlyargs or 'labels' in inspect.getfullargspec(fn).args:
                if 'pos_label' in inspect.getfullargspec(fn).kwonlyargs or 'pos_label' in inspect.getfullargspec(fn).args:
                    score = fn(true_y, predictions, labels=classes, pos_label=pos_class if pos_class is not None else classes[-1])
                else:
                    score = fn(true_y, predictions, labels=classes)
            else:
                score = fn(true_y, predictions)

        except ValueError as err:
            logging.warning(f"MLMethodAssessment: score for metric {metric.name} could not be calculated."
                          f"\nPredicted values: {predicted_y}\nTrue values: {true_y}.\nMore details: {err}")
            score = Constants.NOT_COMPUTED

        return score



import logging
from pathlib import Path
from typing import List, Dict

import numpy as np
import pandas as pd
import plotly.graph_objects as go

from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.environment.Label import Label
from immuneML.hyperparameter_optimization.states.HPItem import HPItem
from immuneML.hyperparameter_optimization.states.TrainMLModelState import TrainMLModelState
from immuneML.ml_metrics.ClassificationMetric import ClassificationMetric
from immuneML.ml_metrics.MetricUtil import MetricUtil
from immuneML.reports.PlotlyUtil import PlotlyUtil
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.train_ml_model_reports.TrainMLModelReport import TrainMLModelReport
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder


class PerformancePerLabel(TrainMLModelReport):
    """
    Report that shows the performance of the model where the examples are grouped by alternative_label. It can be used
    to investigate if the model is learning the alternative_label instead of label of interest for classification.

    **Specification arguments:**

    - alternative_label (str): The name of the alternative_label column in the dataset.

    - metric (str): The metric to use for the report. Default is balanced_accuracy.

    - compute_for_selection (bool): If True, the report will be computed for the selection. Default is True.

    - compute_for_assessment (bool): If True, the report will be computed for the assessment. Default is True.

    - plot_on_test (bool): If True, the report will be plotted on the test data. Default is True.

    - plot_on_train (bool): If True, the report will be plotted on the training data. Default is False.

    **YAML specification:**

    .. code-block:: yaml

        reports:
            my_report:
                PerformancePerLabel:
                    alternative_label: batch
                    metric: balanced_accuracy

    """

    def __init__(self, alternative_label: str, metric: str = "balanced_accuracy", compute_for_selection: bool = True,
                 compute_for_assessment: bool = True, state: TrainMLModelState = None, result_path: Path = None,
                 name: str = None, label: Label = None, number_of_processes: int = 1, plot_on_train: bool = False,
                 plot_on_test: bool = True):
        super().__init__(name=name, state=state, label=label, result_path=result_path,
                         number_of_processes=number_of_processes)
        self.alternative_label = alternative_label
        self.alternative_label_values = []
        self.metric = metric.lower()
        self.compute_for_selection = compute_for_selection
        self.compute_for_assessment = compute_for_assessment
        self.plot_on_train = plot_on_train
        self.plot_on_test = plot_on_test

    @classmethod
    def build_object(cls, **kwargs):
        ParameterValidator.assert_keys_present(list(kwargs.keys()),
                                               ["alternative_label", 'compute_for_selection', 'compute_for_assessment',
                                                'metric'], "PerformancePerLabel",
                                               "PerformancePerLabel")

        alternative_label = kwargs["alternative_label"]
        metric = kwargs["metric"] if "metric" in kwargs else "balanced_accuracy"
        compute_for_selection = kwargs["compute_for_selection"]
        compute_for_assessment = kwargs["compute_for_assessment"]
        name = kwargs["name"] if "name" in kwargs else None

        return PerformancePerLabel(alternative_label=alternative_label, metric=metric,
                                   compute_for_selection=compute_for_selection,
                                   compute_for_assessment=compute_for_assessment,
                                   name=name)

    def discover_alternative_label_values(self):
        self.alternative_label_values = sorted(self.state.dataset.get_metadata([self.alternative_label], return_df=True)[
            self.alternative_label].unique().tolist())

    def _generate(self) -> ReportResult:
        PathBuilder.build(self.result_path)
        report_outputs = []

        self.discover_alternative_label_values()

        if self.compute_for_selection:
            selection_outputs = self._generate_for_selection()
            report_outputs.extend(selection_outputs)

        if self.compute_for_assessment:
            assessment_outputs = self._generate_for_assessment()
            report_outputs.extend(assessment_outputs)

        return ReportResult(name=self.name,
                            info="Performance per label report showing how the model performs on different subgroups of data.",
                            output_tables=[output for output in report_outputs if output.path.suffix == '.csv'],
                            output_figures=[output for output in report_outputs if output.path.suffix == '.html'])

    def _generate_for_selection(self):
        outputs = []
        for assessment_index, assessment_state in enumerate(self.state.assessment_states):
            for label_name, label_state in assessment_state.label_states.items():
                data = self._get_performance_data(label_state.selection_state.hp_items)
                outputs.extend(self._write_split_results(data, f"{label_name}_assessment_{assessment_index+1}_selection", label_name))
        return outputs

    def _generate_for_assessment(self):
        outputs = []
        for label in self.state.label_configuration.get_label_objects():
            assessment_items_per_setting = {hp_setting.get_key(): [] for hp_setting in self.state.hp_settings}

            for assessment_index, assessment_state in enumerate(self.state.assessment_states):
                for hp_setting in self.state.hp_settings:
                    assessment_items_per_setting[hp_setting.get_key()].append(assessment_state.label_states[label.name].assessment_items[hp_setting.get_key()])

            data = self._get_performance_data(assessment_items_per_setting)

            outputs.extend(self._write_split_results(data, f"{label.name}_assessment", label.name))
        return outputs

    def _get_performance_for_dataset(self, predictions: pd.DataFrame, dataset: Dataset, label: Label, counts: dict):
        data = []

        for label_value in self.alternative_label_values:
            mask = dataset.get_metadata([self.alternative_label], return_df=True)[self.alternative_label] == label_value
            indices = np.where(mask)[0]
            if len(indices) > 0:
                true = predictions[f"{label.name}_true_class"].iloc[indices]
                predicted = predictions[f"{label.name}_predicted_class"].iloc[indices]
                if f"{label.name}_{label.positive_class}_proba" in predictions.columns:
                    proba = predictions[f"{label.name}_{label.positive_class}_proba"].iloc[indices]
                else:
                    proba = None
                performance = MetricUtil.score_for_metric(
                    metric=ClassificationMetric[self.metric.upper()],
                    true_y=true,
                    predicted_y=predicted,
                    predicted_proba_y=proba,
                    classes=label.values,
                    pos_class=label.positive_class
                )
            else:
                performance = float('nan')

            data.append({
                "alternative_label_value": label_value,
                "count": counts.get(label_value, 0),
                "performance": performance
            })

        return data

    def _get_performance_data(self, hp_items: Dict[str, List[HPItem]]):
        dfs = {'train': [], 'test': []}

        for label in self.state.label_configuration.get_label_objects():
            for setting_name, hp_item_results in hp_items.items():

                for run_id, hp_item_result in enumerate(hp_item_results):

                    datasets = self._prepare_datasets(hp_item_result)

                    for dataset_info in datasets:
                        if (dataset_info['desc'] == 'train' and self.plot_on_train) or (dataset_info['desc'] == 'test' and self.plot_on_test):
                            data = self._process_dataset(dataset_info, label, hp_item_result)
                            dfs[dataset_info['desc']].append({**data, 'run_id': run_id+1})

        return {'train': pd.DataFrame(dfs['train']) if self.plot_on_train else None,
                'test': pd.DataFrame(dfs['test'] if self.plot_on_test else [])}

    def _prepare_datasets(self, hp_item: HPItem):
        train_metadata = hp_item.train_dataset.get_metadata([self.alternative_label], return_df=True)
        test_metadata = hp_item.test_dataset.get_metadata([self.alternative_label], return_df=True)
        
        train_counts = train_metadata[self.alternative_label].value_counts()
        test_counts = test_metadata[self.alternative_label].value_counts()

        train_predictions = self._load_predictions(hp_item.train_predictions_path)
        test_predictions = self._load_predictions(hp_item.test_predictions_path)

        if train_predictions is None or test_predictions is None:
            self._log_missing_predictions(hp_item)
            return None

        return [
            {'desc': 'train', 'dataset': hp_item.train_dataset, 'predictions': train_predictions, 
             'counts': train_counts, 'metadata': train_metadata},
            {'desc': 'test', 'dataset': hp_item.test_dataset, 'predictions': test_predictions, 
             'counts': test_counts, 'metadata': test_metadata}
        ]

    def _load_predictions(self, path):
        return pd.read_csv(path) if path.is_file() else None

    def _log_missing_predictions(self, hp_item: HPItem):
        logging.warning(f"{self.__class__.__name__}: predictions file not found for split {hp_item.split_index}, "
                        f"encoder {hp_item.hp_setting.encoder_name}, ml_method {hp_item.hp_setting.ml_method_name}")

    def _process_dataset(self, dataset_info: dict, label: Label, hp_item: HPItem):
        predictions = dataset_info['predictions']
        metadata = dataset_info['metadata']
        
        # Calculate overall performance
        overall_performance = self._calculate_performance(predictions, label)
        
        # Calculate per-label performances and counts
        alt_label_performances = {}
        alt_label_counts = {}
        total_count = len(predictions)
        
        for alt_label_value in self.alternative_label_values:
            perf = self._get_performance_for_label_value(predictions, metadata, label, alt_label_value)
            count = self._get_count_for_label_value(metadata, alt_label_value)
            
            alt_label_performances[f"performance_{alt_label_value}"] = perf
            alt_label_counts[f"count_{alt_label_value}"] = count
        
        return {
            "setting": hp_item.hp_setting.get_key(),
            "performance": overall_performance,
            "example_count": total_count,
            **alt_label_performances,
            **alt_label_counts
        }

    def _get_performance_for_label_value(self, predictions, metadata, label, alt_label_value):
        mask = metadata[self.alternative_label] == alt_label_value
        indices = np.where(mask)[0]
        
        if len(indices) > 0:
            true = predictions[f"{label.name}_true_class"].iloc[indices]
            predicted = predictions[f"{label.name}_predicted_class"].iloc[indices]
            proba = self._get_probabilities(predictions, label)
            if proba is not None:
                proba = proba.iloc[indices]
            
            return self._compute_metric(true, predicted, proba, label)
        return float('nan')

    def _get_count_for_label_value(self, metadata, alt_label_value):
        return len(metadata[metadata[self.alternative_label] == alt_label_value])

    def _get_probabilities(self, predictions, label):
        proba_col = f"{label.name}_{label.positive_class}_proba"
        return predictions[proba_col] if proba_col in predictions.columns else None

    def _calculate_performance(self, predictions, label):
        true = predictions[f"{label.name}_true_class"]
        predicted = predictions[f"{label.name}_predicted_class"]
        proba = self._get_probabilities(predictions, label)
        
        return self._compute_metric(true, predicted, proba, label)

    def _compute_metric(self, true, predicted, proba, label):
        return MetricUtil.score_for_metric(
            metric=ClassificationMetric[self.metric.upper()],
            true_y=true,
            predicted_y=predicted,
            predicted_proba_y=proba,
            classes=label.values,
            pos_class=label.positive_class
        )

    def _write_split_results(self, data: Dict[str, pd.DataFrame], name_suffix: str, label_name: str):
        outputs = []
        datasets = []
        if self.plot_on_train:
            datasets.append('train')
        if self.plot_on_test:
            datasets.append('test')

        for dataset_desc in datasets:
            outputs.append(self._write_performance_tables(data[dataset_desc], dataset_desc, name_suffix + "_" + dataset_desc, label_name))
            outputs.append(self._create_performance_plot(data[dataset_desc], dataset_desc, name_suffix + "_" + dataset_desc, label_name))

        outputs = [x for item in outputs for x in (item if isinstance(item, list) else [item])]

        return outputs

    def _write_performance_tables(self, data: pd.DataFrame, dataset_desc: str, name_suffix: str, label_name: str):
        table_path = self.result_path / f"{name_suffix}_performance.csv"
        data.to_csv(table_path, index=False)
        return ReportOutput(table_path, self._get_desc_from_name_suffix(name_suffix, label_name, dataset_desc))

    def _get_desc_from_name_suffix(self, name_suffix: str, label_name: str, dataset_desc: str):
        return (f"Performance on {label_name} split by {self.alternative_label} ({dataset_desc} set "
                f"in {'-'.join([el for el in name_suffix.replace(label_name, '').replace(dataset_desc, '').split('_') if el != ''])})")

    def _reorder_columns(self, df):
        # Base columns
        cols = ["setting", "performance", "example_count"]
        
        # Add performance columns for each alternative label value
        perf_cols = [f"performance_{val}" for val in self.alternative_label_values]
        
        # Add count columns for each alternative label value
        count_cols = [f"count_{val}" for val in self.alternative_label_values]
        
        return df[cols + perf_cols + count_cols]

    def _create_performance_plot(self, data: pd.DataFrame, dataset_desc: str, name_suffix, label_name: str):
        fig = self._create_figure(data, label_name)
        plot_path = self.result_path / f"{name_suffix}_performance_plot.html"
        plot_path = PlotlyUtil.write_image_to_file(fig, plot_path)
        return ReportOutput(plot_path,
                            self._get_desc_from_name_suffix(name_suffix, label_name, dataset_desc))

    def _create_figure(self, data: pd.DataFrame, label_name: str):
        fig = go.Figure()
        groups_for_perf_eval = ['all'] + self.alternative_label_values
        repetitions = data.run_id.unique().shape[0]
        groups_for_perf_eval = np.repeat(groups_for_perf_eval, repetitions)

        for setting in data.setting.unique().tolist():
            setting_data = data[data.setting == setting]
            y = np.concatenate([setting_data.performance] + [setting_data[f'performance_{alt_lbl_value}'] for alt_lbl_value in self.alternative_label_values]).round(3)
            if repetitions > 1:
                fig.add_trace(go.Box(
                    name=setting,
                    x=groups_for_perf_eval,
                    y=y,
                    boxpoints='all',
                    jitter=0.3
                ))
            else:
                fig.add_trace(go.Bar(
                    name=setting,
                    x=groups_for_perf_eval,
                    y=y,
                    hovertemplate=setting + "<br>" + self.alternative_label + ": %{x}<br>" + self.metric + ": %{y}<extra></extra>"
                ))
        
        fig.update_layout(**self._get_layout_settings({'boxmode': 'group'} if repetitions > 1 else {'barmode': 'group'}))
        return fig

    def _get_layout_settings(self, kwargs):
        return {
            **kwargs,
            "title": f"Performance by {self.alternative_label}",
            "xaxis_title": self.alternative_label,
            "yaxis_title": f"{self.metric.replace('_', ' ').title()}",
            "template": "plotly_white",
            "showlegend": True
        }

    def check_prerequisites(self):
        if self.state is None:
            logging.warning(
                f"{self.__class__.__name__} requires state to be set. {self.__class__.__name__} report will not be created.")
            return False
        if not self.compute_for_selection and not self.compute_for_assessment:
            logging.warning(
                f"{self.__class__.__name__} requires either compute_for_selection or compute_for_assessment to be True. "
                f"{self.__class__.__name__} report will not be created.")
            return False
        return True
